# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import os
import os.path
import unittest

import mock
import requests
import requests_unixsocket

from pylxd import client, exceptions
from pylxd.tests.testing import requires_ws4py


class TestClient(unittest.TestCase):
    """Tests for pylxd.client.Client."""

    def setUp(self):
        self.get_patcher = mock.patch('pylxd.client._APINode.get')
        self.get = self.get_patcher.start()

        self.post_patcher = mock.patch('pylxd.client._APINode.post')
        self.post = self.post_patcher.start()

        response = mock.MagicMock(status_code=200)
        response.json.return_value = {'metadata': {
            'auth': 'trusted',
            'environment': {'storage': 'zfs'},
        }}
        self.get.return_value = response

        post_response = mock.MagicMock(status_code=200)
        self.post.return_value = post_response

    def tearDown(self):
        self.get_patcher.stop()
        self.post_patcher.stop()

    def test_create(self):
        """Client creation sets default API endpoint."""
        expected = 'http+unix://%2Fvar%2Flib%2Flxd%2Funix.socket/1.0'

        an_client = client.Client()

        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_create_LXD_DIR(self):
        """When LXD_DIR is set, use it in the client."""
        os.environ['LXD_DIR'] = '/lxd'
        expected = 'http+unix://%2Flxd%2Funix.socket/1.0'

        an_client = client.Client()

        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_create_endpoint(self):
        """Explicitly set the client endpoint."""
        endpoint = 'http://lxd'
        expected = 'http://lxd/1.0'

        an_client = client.Client(endpoint=endpoint)

        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_create_endpoint_unixsocket(self):
        """Test with unix socket endpoint."""
        endpoint = '/tmp/unix.socket'
        expected = 'http+unix://%2Ftmp%2Funix.socket/1.0'

        real_isfile = os.path.isfile
        os.path.isfile = lambda x: True
        an_client = client.Client(endpoint)
        os.path.isfile = real_isfile

        self.assertEqual(expected, an_client.api._api_endpoint)

    def test_connection_404(self):
        """If the endpoint 404s, an exception is raised."""
        response = mock.MagicMock(status_code=404)
        self.get.return_value = response

        self.assertRaises(exceptions.ClientConnectionFailed, client.Client)

    def test_connection_failed(self):
        """If the connection fails, an exception is raised."""
        def raise_exception():
            raise requests.exceptions.ConnectionError()
        self.get.side_effect = raise_exception
        self.get.return_value = None

        self.assertRaises(exceptions.ClientConnectionFailed, client.Client)

    def test_connection_untrusted(self):
        """Client.trusted is False when certs are untrusted."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {'metadata': {'auth': 'untrusted'}}
        self.get.return_value = response

        an_client = client.Client()

        self.assertFalse(an_client.trusted)

    def test_connection_trusted(self):
        """Client.trusted is True when certs are untrusted."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {'metadata': {'auth': 'trusted'}}
        self.get.return_value = response

        an_client = client.Client()

        self.assertTrue(an_client.trusted)

    def test_authenticate(self):
        """A client is authenticated."""
        response = mock.MagicMock(status_code=200)
        response.json.return_value = {'metadata': {'auth': 'untrusted'}}
        self.get.return_value = response

        certs = (
            os.path.join(os.path.dirname(__file__), 'lxd.crt'),
            os.path.join(os.path.dirname(__file__), 'lxd.key'))
        an_client = client.Client('https://lxd', cert=certs)

        get_count = []

        def _get(*args, **kwargs):
            if len(get_count) == 0:
                get_count.append(None)
                return {'metadata': {
                    'type': 'client',
                    'fingerprint': 'eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c',  # NOQA
                }}
            else:
                return {'metadata': {'auth': 'trusted'}}
        response = mock.MagicMock(status_code=200)
        response.json.side_effect = _get
        self.get.return_value = response

        an_client.authenticate('test-password')

        self.assertTrue(an_client.trusted)

    def test_authenticate_already_authenticated(self):
        """If the client is already authenticated, nothing happens."""
        an_client = client.Client()

        an_client.authenticate('test-password')

        self.assertTrue(an_client.trusted)

    def test_host_info(self):
        """Perform a host query."""
        an_client = client.Client()
        self.assertEqual('zfs', an_client.host_info['environment']['storage'])

    @requires_ws4py
    def test_events(self):
        """The default websocket client is returned."""
        an_client = client.Client()

        ws_client = an_client.events()

        self.assertEqual('/1.0/events', ws_client.resource)

    def test_events_no_ws4py(self):
        """No ws4py will result in a ValueError."""
        from pylxd import client
        old_installed = client._ws4py_installed
        client._ws4py_installed = False

        def cleanup():
            client._ws4py_installed = old_installed
        self.addCleanup(cleanup)

        an_client = client.Client()

        self.assertRaises(ValueError, an_client.events)
        client._ws4py_installed

    @requires_ws4py
    def test_events_unix_socket(self):
        """A unix socket compatible websocket client is returned."""
        websocket_client = mock.Mock(resource=None)
        WebsocketClient = mock.Mock()
        WebsocketClient.return_value = websocket_client
        an_client = client.Client()

        an_client.events(websocket_client=WebsocketClient)

        WebsocketClient.assert_called_once_with('ws+unix:///lxd/unix.socket')

    @requires_ws4py
    def test_events_htt(self):
        """An http compatible websocket client is returned."""
        websocket_client = mock.Mock(resource=None)
        WebsocketClient = mock.Mock()
        WebsocketClient.return_value = websocket_client
        an_client = client.Client('http://lxd.local')

        an_client.events(websocket_client=WebsocketClient)

        WebsocketClient.assert_called_once_with('ws://lxd.local')

    @requires_ws4py
    def test_events_https(self):
        """An https compatible websocket client is returned."""
        websocket_client = mock.Mock(resource=None)
        WebsocketClient = mock.Mock()
        WebsocketClient.return_value = websocket_client
        an_client = client.Client('https://lxd.local')

        an_client.events(websocket_client=WebsocketClient)

        WebsocketClient.assert_called_once_with('wss://lxd.local')


class TestAPINode(unittest.TestCase):
    """Tests for pylxd.client._APINode."""

    def test_getattr(self):
        """API Nodes can use object notation for nesting."""
        node = client._APINode('http://test.com')

        new_node = node.test

        self.assertEqual(
            'http://test.com/test', new_node._api_endpoint)

    def test_getitem(self):
        """API Nodes can use dict notation for nesting."""
        node = client._APINode('http://test.com')

        new_node = node['test']

        self.assertEqual(
            'http://test.com/test', new_node._api_endpoint)

    def test_session_http(self):
        """HTTP nodes return the default requests session."""
        node = client._APINode('http://test.com')

        self.assertIsInstance(node.session, requests.Session)

    def test_session_unix_socket(self):
        """HTTP nodes return a requests_unixsocket session."""
        node = client._APINode('http+unix://test.com')

        self.assertIsInstance(node.session, requests_unixsocket.Session)

    @mock.patch('pylxd.client.requests.Session')
    def test_get(self, Session):
        """Perform a session get."""
        response = mock.Mock(**{
            'status_code': 200,
            'json.return_value': {'type': 'sync'},
        })
        session = mock.Mock(**{'get.return_value': response})
        Session.return_value = session

        node = client._APINode('http://test.com')

        node.get()

        session.get.assert_called_once_with('http://test.com')

    @mock.patch('pylxd.client.requests.Session')
    def test_post(self, Session):
        """Perform a session post."""
        response = mock.Mock(**{
            'status_code': 200,
            'json.return_value': {'type': 'sync'},
        })
        session = mock.Mock(**{'post.return_value': response})
        Session.return_value = session

        node = client._APINode('http://test.com')

        node.post()

        session.post.assert_called_once_with('http://test.com')

    @mock.patch('pylxd.client.requests.Session')
    def test_post_200_not_sync(self, Session):
        """A status code of 200 with async request raises an exception."""
        response = mock.Mock(**{
            'status_code': 200,
            'json.return_value': {'type': 'async'},
        })
        session = mock.Mock(**{'post.return_value': response})
        Session.return_value = session

        node = client._APINode('http://test.com')

        self.assertRaises(
            exceptions.LXDAPIException,
            node.post)

    @mock.patch('pylxd.client.requests.Session')
    def test_post_missing_type_200(self, Session):
        """A missing response type raises an exception."""
        response = mock.Mock(**{
            'status_code': 200,
            'json.return_value': {},
        })
        session = mock.Mock(**{'post.return_value': response})
        Session.return_value = session

        node = client._APINode('http://test.com')

        self.assertRaises(
            exceptions.LXDAPIException,
            node.post)

    @mock.patch('pylxd.client.requests.Session')
    def test_put(self, Session):
        """Perform a session put."""
        response = mock.Mock(**{
            'status_code': 200,
            'json.return_value': {'type': 'sync'},
        })
        session = mock.Mock(**{'put.return_value': response})
        Session.return_value = session

        node = client._APINode('http://test.com')

        node.put()

        session.put.assert_called_once_with('http://test.com')

    @mock.patch('pylxd.client.requests.Session')
    def test_delete(self, Session):
        """Perform a session delete."""
        response = mock.Mock(**{
            'status_code': 200,
            'json.return_value': {'type': 'sync'},
        })
        session = mock.Mock(**{'delete.return_value': response})
        Session.return_value = session

        node = client._APINode('http://test.com')

        node.delete()

        session.delete.assert_called_once_with('http://test.com')


class TestWebsocketClient(unittest.TestCase):
    """Tests for pylxd.client.WebsocketClient."""

    @requires_ws4py
    def test_handshake_ok(self):
        """A `message` attribute of an empty list is created."""
        ws_client = client._WebsocketClient('ws://an/fake/path')

        ws_client.handshake_ok()

        self.assertEqual([], ws_client.messages)

    @requires_ws4py
    def test_received_message(self):
        """A json dict is added to the messages attribute."""
        message = mock.Mock(data=json.dumps({'test': 'data'}).encode('utf-8'))
        ws_client = client._WebsocketClient('ws://an/fake/path')
        ws_client.handshake_ok()

        ws_client.received_message(message)

        self.assertEqual({'test': 'data'}, ws_client.messages[0])



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from pylxd.models import _model as model
from pylxd.tests import testing


class Item(model.Model):
    """A fake model."""
    name = model.Attribute(readonly=True)
    age = model.Attribute(int)
    data = model.Attribute()

    @property
    def api(self):
        return self.client.api.items[self.name]


class TestModel(testing.PyLXDTestCase):
    """Tests for pylxd.model.Model."""

    def setUp(self):
        super(TestModel, self).setUp()

        self.add_rule({
            'json': {
                'type': 'sync',
                'metadata': {
                    'name': 'an-item',
                    'age': 1000,
                    'data': {'key': 'val'},
                }
            },
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/items/an-item',
        })
        self.add_rule({
            'json': {
                'type': 'sync',
                'metadata': {}
            },
            'method': 'PUT',
            'url': r'^http://pylxd.test/1.0/items/an-item',
        })
        self.add_rule({
            'json': {
                'type': 'sync',
                'metadata': {}
            },
            'method': 'DELETE',
            'url': r'^http://pylxd.test/1.0/items/an-item',
        })

    def test_init(self):
        """Initial attributes are set."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        self.assertEqual(self.client, item.client)
        self.assertEqual('an-item', item.name)

    def test_init_unknown_attribute(self):
        """Unknown attributes aren't set."""
        item = Item(self.client, name='an-item', nonexistent='SRSLY')

        try:
            item.nonexistent
            self.fail('item.nonexistent did not raise AttributeError')
        except AttributeError:
            pass

    def test_unknown_attribute(self):
        """Setting unknown attributes raise an exception."""
        def set_unknown_attribute():
            item = Item(self.client, name='an-item')
            item.nonexistent = 'SRSLY'
        self.assertRaises(AttributeError, set_unknown_attribute)

    def test_get_unknown_attribute(self):
        """Setting unknown attributes raise an exception."""
        def get_unknown_attribute():
            item = Item(self.client, name='an-item')
            return item.nonexistent
        self.assertRaises(AttributeError, get_unknown_attribute)

    def test_unset_attribute_sync(self):
        """Reading unavailable attributes calls sync."""
        item = Item(self.client, name='an-item')

        self.assertEqual(1000, item.age)

    def test_sync(self):
        """A sync will update attributes from the server."""
        item = Item(self.client, name='an-item')

        item.sync()

        self.assertEqual(1000, item.age)

    def test_sync_dirty(self):
        """Sync will not overwrite local attribute changes."""
        item = Item(self.client, name='an-item')

        item.age = 250
        item.sync()

        self.assertEqual(250, item.age)

    def test_rollback(self):
        """Rollback resets the object from the server."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        item.age = 50
        item.rollback()

        self.assertEqual(1000, item.age)
        self.assertFalse(item.dirty)

    def test_int_attribute_validator(self):
        """Age is set properly to be an int."""
        item = Item(self.client)

        item.age = '100'

        self.assertEqual(100, item.age)

    def test_int_attribute_invalid(self):
        """TypeError is raised when data can't be converted to type."""
        def set_string():
            item = Item(self.client)
            item.age = 'abc'

        self.assertRaises(ValueError, set_string)

    def test_dirty(self):
        """Changes mark the object as dirty."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        item.age = 100

        self.assertTrue(item.dirty)

    def test_not_dirty(self):
        """Changes mark the object as dirty."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        self.assertFalse(item.dirty)

    def test_marshall(self):
        """The object is marshalled into a dict."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        result = item.marshall()

        self.assertEqual({'age': 15, 'data': {'key': 'val'}}, result)

    def test_delete(self):
        """The object is deleted, and client is unset."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        item.delete()

        self.assertIsNone(item.client)

    def test_save(self):
        """Attributes are written to the server; object is marked clean."""
        item = Item(self.client, name='an-item', age=15, data={'key': 'val'})

        item.age = 69
        item.save()

        self.assertFalse(item.dirty)



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ddt import data
from ddt import ddt
import mock

from pylxd.deprecated import connection

from pylxd.deprecated.tests import annotated_data
from pylxd.deprecated.tests import fake_api
from pylxd.deprecated.tests import LXDAPITestBase


@ddt
@mock.patch.object(connection.LXDConnection, 'get_object')
class LXDAPIImageAliasTestObject(LXDAPITestBase):

    def test_alias_list(self, ms):
        ms.return_value = ('200', fake_api.fake_alias_list())
        self.assertEqual(['ubuntu'], self.lxd.alias_list())
        ms.assert_called_once_with('GET', '/1.0/images/aliases')

    def test_alias_show(self, ms):
        ms.return_value = ('200', fake_api.fake_alias())
        self.assertEqual(
            fake_api.fake_alias(), self.lxd.alias_show('fake')[1])
        ms.assert_called_once_with('GET', '/1.0/images/aliases/fake')


@ddt
@mock.patch.object(connection.LXDConnection, 'get_status')
class LXDAPIImageAliasTestStatus(LXDAPITestBase):

    @data(True, False)
    def test_alias_defined(self, expected, ms):
        ms.return_value = expected
        self.assertEqual(expected, self.lxd.alias_defined('fake'))
        ms.assert_called_once_with('GET', '/1.0/images/aliases/fake')

    @annotated_data(
        ('create', 'POST', '', ('fake',), ('"fake"',)),
        ('update', 'PUT', '/test-alias',
         ('test-alias', 'fake',), ('"fake"',)),
        ('rename', 'POST', '/test-alias',
         ('test-alias', 'fake',), ('"fake"',)),
        ('delete', 'DELETE', '/test-alias', ('test-alias',), ()),
    )
    def test_alias_operations(self, method, http, path, args, call_args, ms):
        self.assertTrue(getattr(self.lxd, 'alias_' + method)(*args))
        ms.assert_called_once_with(
            http,
            '/1.0/images/aliases' + path,
            *call_args
        )



import hashlib
import json

from pylxd import exceptions, models
from pylxd.tests import testing


class TestImage(testing.PyLXDTestCase):
    """Tests for pylxd.models.Image."""

    def test_get(self):
        """An image is fetched."""
        fingerprint = hashlib.sha256(b'').hexdigest()
        a_image = models.Image.get(self.client, fingerprint)

        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_get_not_found(self):
        """LXDAPIException is raised when the image isn't found."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
        })

        fingerprint = hashlib.sha256(b'').hexdigest()

        self.assertRaises(
            exceptions.LXDAPIException,
            models.Image.get, self.client, fingerprint)

    def test_get_error(self):
        """LXDAPIException is raised on error."""
        def error(request, context):
            context.status_code = 500
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 500})
        self.add_rule({
            'text': error,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
        })

        fingerprint = hashlib.sha256(b'').hexdigest()

        self.assertRaises(
            exceptions.LXDAPIException,
            models.Image.get, self.client, fingerprint)

    def test_get_by_alias(self):
        fingerprint = hashlib.sha256(b'').hexdigest()

        a_image = models.Image.get_by_alias(self.client, 'an-alias')

        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_exists(self):
        """An image is fetched."""
        fingerprint = hashlib.sha256(b'').hexdigest()

        self.assertTrue(models.Image.exists(self.client, fingerprint))

    def test_exists_by_alias(self):
        """An image is fetched."""
        self.assertTrue(models.Image.exists(
            self.client, 'an-alias', alias=True))

    def test_not_exists(self):
        """LXDAPIException is raised when the image isn't found."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
        })

        fingerprint = hashlib.sha256(b'').hexdigest()

        self.assertFalse(models.Image.exists(self.client, fingerprint))

    def test_all(self):
        """A list of all images is returned."""
        images = models.Image.all(self.client)

        self.assertEqual(1, len(images))

    def test_create(self):
        """An image is created."""
        fingerprint = hashlib.sha256(b'').hexdigest()
        a_image = models.Image.create(
            self.client, b'', public=True, wait=True)

        self.assertIsInstance(a_image, models.Image)
        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_create_with_metadata(self):
        """An image with metadata is created."""
        fingerprint = hashlib.sha256(b'').hexdigest()
        a_image = models.Image.create(
            self.client, b'', metadata=b'', public=True, wait=True)

        self.assertIsInstance(a_image, models.Image)
        self.assertEqual(fingerprint, a_image.fingerprint)

    def test_update(self):
        """An image is updated."""
        a_image = self.client.images.all()[0]
        a_image.sync()

        a_image.save()

    def test_fetch(self):
        """A partial object is fetched and populated."""
        a_image = self.client.images.all()[0]

        a_image.sync()

        self.assertEqual(1, a_image.size)

    def test_fetch_notfound(self):
        """A bogus image fetch raises LXDAPIException."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
        })
        fingerprint = hashlib.sha256(b'').hexdigest()

        a_image = models.Image(self.client, fingerprint=fingerprint)

        self.assertRaises(exceptions.LXDAPIException, a_image.sync)

    def test_fetch_error(self):
        """A 500 error raises LXDAPIException."""
        def not_found(request, context):
            context.status_code = 500
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 500})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
        })
        fingerprint = hashlib.sha256(b'').hexdigest()

        a_image = models.Image(self.client, fingerprint=fingerprint)

        self.assertRaises(exceptions.LXDAPIException, a_image.sync)

    def test_delete(self):
        """An image is deleted."""
        # XXX: rockstar (03 Jun 2016) - This just executes
        # a code path. There should be an assertion here, but
        # it's not clear how to assert that, just yet.
        a_image = self.client.images.all()[0]

        a_image.delete(wait=True)

    def test_export(self):
        """An image is exported."""
        expected = 'e2943f8d0b0e7d5835f9533722a6e25f669acb8980daee378b4edb44da212f51'  # NOQA
        a_image = self.client.images.all()[0]

        data = a_image.export()
        data_sha = hashlib.sha256(data.read()).hexdigest()

        self.assertEqual(expected, data_sha)

    def test_export_not_found(self):
        """LXDAPIException is raised on export of bogus image."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/export$',  # NOQA
        })
        a_image = self.client.images.all()[0]

        self.assertRaises(exceptions.LXDAPIException, a_image.export)

    def test_export_error(self):
        """LXDAPIException is raised on API error."""
        def error(request, context):
            context.status_code = 500
            return json.dumps({
                'type': 'error',
                'error': 'LOLOLOLOL',
                'error_code': 500})
        self.add_rule({
            'text': error,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/export$',  # NOQA
        })
        a_image = self.client.images.all()[0]

        self.assertRaises(exceptions.LXDAPIException, a_image.export)

    def test_add_alias(self):
        """Try to add an alias."""
        a_image = self.client.images.all()[0]
        a_image.add_alias('lol', 'Just LOL')

        aliases = [a['name'] for a in a_image.aliases]
        self.assertTrue('lol' in aliases, "Image didn't get updated.")

    def test_add_alias_duplicate(self):
        """Adding a alias twice should raise an LXDAPIException."""
        def error(request, context):
            context.status_code = 409
            return json.dumps({
                'type': 'error',
                'error': 'already exists',
                'error_code': 409})
        self.add_rule({
            'text': error,
            'method': 'POST',
            'url': r'^http://pylxd.test/1.0/images/aliases$',  # NOQA
        })

        a_image = self.client.images.all()[0]

        self.assertRaises(
            exceptions.LXDAPIException,
            a_image.add_alias,
            'lol', 'Just LOL'
        )

    def test_remove_alias(self):
        """Try to remove an-alias."""
        a_image = self.client.images.all()[0]
        a_image.delete_alias('an-alias')

        self.assertEqual(0, len(a_image.aliases), "Alias didn't get deleted.")

    def test_remove_alias_error(self):
        """Try to remove an non existant alias."""
        def error(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'not found',
                'error_code': 404})
        self.add_rule({
            'text': error,
            'method': 'DELETE',
            'url': r'^http://pylxd.test/1.0/images/aliases/lol$',  # NOQA
        })

        a_image = self.client.images.all()[0]
        self.assertRaises(
            exceptions.LXDAPIException,
            a_image.delete_alias,
            'lol'
        )

    def test_remove_alias_not_in_image(self):
        """Try to remove an alias which is not in the current image."""
        a_image = self.client.images.all()[0]
        a_image.delete_alias('b-alias')

    def test_copy(self):
        """Try to copy an image to another LXD instance."""
        from pylxd.client import Client

        a_image = self.client.images.all()[0]

        client2 = Client(endpoint='http://pylxd2.test')
        copied_image = a_image.copy(client2, wait=True)
        self.assertEqual(a_image.fingerprint, copied_image.fingerprint)

    def test_copy_public(self):
        """Try to copy a public image."""
        from pylxd.client import Client

        def image_get(request, context):
            context.status_code = 200
            return json.dumps({
                'type': 'sync',
                'metadata': {
                    'aliases': [
                        {
                            'name': 'an-alias',  # NOQA
                            'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
                        }
                    ],
                    'architecture': 'x86_64',
                    'cached': False,
                    'filename': 'a_image.tar.bz2',
                    'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
                    'public': True,
                    'properties': {},
                    'size': 1,
                    'auto_update': False,
                    'created_at': '1983-06-16T02:42:00Z',
                    'expires_at': '1983-06-16T02:42:00Z',
                    'last_used_at': '1983-06-16T02:42:00Z',
                    'uploaded_at': '1983-06-16T02:42:00Z',

                },
            })
        self.add_rule({
            'text': image_get,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
        })

        a_image = self.client.images.all()[0]
        self.assertTrue(a_image.public)

        client2 = Client(endpoint='http://pylxd2.test')
        copied_image = a_image.copy(client2, wait=True)
        self.assertEqual(a_image.fingerprint, copied_image.fingerprint)

    def test_copy_no_wait(self):
        """Try to copy and don't wait."""
        from pylxd.client import Client

        a_image = self.client.images.all()[0]

        client2 = Client(endpoint='http://pylxd2.test')
        a_image.copy(client2, public=False, auto_update=False)

    def test_create_from_simplestreams(self):
        """Try to create an image from simplestreams."""
        image = self.client.images.create_from_simplestreams(
            'https://cloud-images.ubuntu.com/releases',
            'trusty/amd64'
        )
        self.assertEqual(
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            image.fingerprint
        )

    def test_create_from_url(self):
        """Try to create an image from an URL."""
        image = self.client.images.create_from_url(
            'https://dl.stgraber.org/lxd'
        )
        self.assertEqual(
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            image.fingerprint
        )



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pylxd.deprecated import base


class LXDNetwork(base.LXDBase):

    def network_list(self):
        (state, data) = self.connection.get_object('GET', '/1.0/networks')
        return [network.split('/1.0/networks/')[-1]
                for network in data['metadata']]

    def network_show(self, network):
        '''Show details of the LXD network'''
        (state, data) = self.connection.get_object('GET', '/1.0/networks/%s'
                                                   % network)
        return {
            'network_name':
                self.show_network_name(network, data.get('metadata')),
            'network_type':
                self.show_network_type(network, data.get('metadata')),
            'network_members':
                self.show_network_members(network, data.get('metadata'))
        }

    def show_network_name(self, network, data):
        '''Show the LXD network name'''
        if data is None:
            (state, data) = self.connection.get_object(
                'GET', '/1.0/networks/%s' % network)
            data = data.get('metadata')
        return data['name']

    def show_network_type(self, network, data):
        if data is None:
            (state, data) = self.connection.get_object(
                'GET', '/1.0/networks/%s' % network)
            data = data.get('metadata')
        return data['type']

    def show_network_members(self, network, data):
        if data is None:
            (state, data) = self.connection.get_object(
                'GET', '/1.0/networks/%s' % network)
            data = data.get('metadata')
        return [network_members.split('/1.0/networks/')[-1]
                for network_members in data['members']]



import json

import mock

from pylxd import exceptions, models
from pylxd.tests import testing


class TestContainer(testing.PyLXDTestCase):
    """Tests for pylxd.models.Container."""

    def test_all(self):
        """A list of all containers are returned."""
        containers = models.Container.all(self.client)

        self.assertEqual(1, len(containers))

    def test_get(self):
        """Return a container."""
        name = 'an-container'

        an_container = models.Container.get(self.client, name)

        self.assertEqual(name, an_container.name)

    def test_get_not_found(self):
        """LXDAPIException is raised when the container doesn't exist."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-missing-container$',  # NOQA
        })

        name = 'an-missing-container'

        self.assertRaises(
            exceptions.LXDAPIException,
            models.Container.get, self.client, name)

    def test_get_error(self):
        """LXDAPIException is raised when the LXD API errors."""
        def not_found(request, context):
            context.status_code = 500
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 500})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-missing-container$',  # NOQA
        })

        name = 'an-missing-container'

        self.assertRaises(
            exceptions.LXDAPIException,
            models.Container.get, self.client, name)

    def test_create(self):
        """A new container is created."""
        config = {'name': 'an-new-container'}

        an_new_container = models.Container.create(
            self.client, config, wait=True)

        self.assertEqual(config['name'], an_new_container.name)

    def test_exists(self):
        """A container exists."""
        name = 'an-container'

        self.assertTrue(models.Container.exists(self.client, name))

    def test_not_exists(self):
        """A container exists."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-missing-container$',  # NOQA
        })

        name = 'an-missing-container'

        self.assertFalse(models.Container.exists(self.client, name))

    def test_fetch(self):
        """A sync updates the properties of a container."""
        an_container = models.Container(
            self.client, name='an-container')

        an_container.sync()

        self.assertTrue(an_container.ephemeral)

    def test_fetch_not_found(self):
        """LXDAPIException is raised on a 404 for updating container."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-missing-container$',  # NOQA
        })

        an_container = models.Container(
            self.client, name='an-missing-container')

        self.assertRaises(exceptions.LXDAPIException, an_container.sync)

    def test_fetch_error(self):
        """LXDAPIException is raised on error."""
        def not_found(request, context):
            context.status_code = 500
            return json.dumps({
                'type': 'error',
                'error': 'An bad error',
                'error_code': 500})
        self.add_rule({
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-missing-container$',  # NOQA
        })

        an_container = models.Container(
            self.client, name='an-missing-container')

        self.assertRaises(exceptions.LXDAPIException, an_container.sync)

    def test_update(self):
        """A container is updated."""
        an_container = models.Container(
            self.client, name='an-container')
        an_container.architecture = 1
        an_container.config = {}
        an_container.created_at = 1
        an_container.devices = {}
        an_container.ephemeral = 1
        an_container.expanded_config = {}
        an_container.expanded_devices = {}
        an_container.profiles = 1
        an_container.status = 1

        an_container.save(wait=True)

        self.assertTrue(an_container.ephemeral)

    def test_rename(self):
        an_container = models.Container(
            self.client, name='an-container')

        an_container.rename('an-renamed-container', wait=True)

        self.assertEqual('an-renamed-container', an_container.name)

    def test_delete(self):
        """A container is deleted."""
        # XXX: rockstar (21 May 2016) - This just executes
        # a code path. There should be an assertion here, but
        # it's not clear how to assert that, just yet.
        an_container = models.Container(
            self.client, name='an-container')

        an_container.delete(wait=True)

    @testing.requires_ws4py
    @mock.patch('pylxd.models.container._StdinWebsocket')
    @mock.patch('pylxd.models.container._CommandWebsocketClient')
    def test_execute(self, _CommandWebsocketClient, _StdinWebsocket):
        """A command is executed on a container."""
        fake_websocket = mock.Mock()
        fake_websocket.data = 'test\n'
        _StdinWebsocket.return_value = fake_websocket
        _CommandWebsocketClient.return_value = fake_websocket

        an_container = models.Container(
            self.client, name='an-container')

        result = an_container.execute(['echo', 'test'])

        self.assertEqual(0, result.exit_code)
        self.assertEqual('test\n', result.stdout)

    def test_execute_no_ws4py(self):
        """If ws4py is not installed, ValueError is raised."""
        from pylxd.models import container
        old_installed = container._ws4py_installed
        container._ws4py_installed = False

        def cleanup():
            container._ws4py_installed = old_installed
        self.addCleanup(cleanup)

        an_container = models.Container(
            self.client, name='an-container')

        self.assertRaises(ValueError, an_container.execute, ['echo', 'test'])

    @testing.requires_ws4py
    def test_execute_string(self):
        """A command passed as string raises a TypeError."""
        an_container = models.Container(
            self.client, name='an-container')

        self.assertRaises(TypeError, an_container.execute, 'apt-get update')

    def test_migrate(self):
        """A container is migrated."""
        from pylxd.client import Client

        client2 = Client(endpoint='http://pylxd2.test')
        an_container = models.Container(
            self.client, name='an-container')

        an_migrated_container = an_container.migrate(client2)

        self.assertEqual('an-container', an_migrated_container.name)
        self.assertEqual(client2, an_migrated_container.client)

    @mock.patch('pylxd.client._APINode.get')
    def test_migrate_local_client(self, get):
        """Migration from local clients is not supported."""
        # Mock out the _APINode for the local instance.
        response = mock.Mock()
        response.json.return_value = {'metadata': {'fake': 'response'}}
        response.status_code = 200
        get.return_value = response

        from pylxd.client import Client

        client2 = Client(endpoint='http+unix://pylxd2.test')
        an_container = models.Container(
            client2, name='an-container')

        self.assertRaises(ValueError, an_container.migrate, self.client)

    def test_publish(self):
        """Containers can be published."""
        self.add_rule({
            'text': json.dumps({
                'type': 'sync',
                'metadata': {
                    'id': 'operation-abc',
                    'metadata': {
                        'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'  # NOQA
                        }
                    }
                }),
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/operations/operation-abc$',
        })

        an_container = models.Container(
            self.client, name='an-container')

        image = an_container.publish(wait=True)

        self.assertEqual(
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            image.fingerprint)


class TestContainerState(testing.PyLXDTestCase):
    """Tests for pylxd.models.ContainerState."""

    def test_get(self):
        """Return a container."""
        name = 'an-container'

        an_container = models.Container.get(self.client, name)
        state = an_container.state()

        self.assertEqual('Running', state.status)
        self.assertEqual(103, state.status_code)

    def test_start(self):
        """A container is started."""
        an_container = models.Container.get(self.client, 'an-container')

        an_container.start(wait=True)

    def test_stop(self):
        """A container is stopped."""
        an_container = models.Container.get(self.client, 'an-container')

        an_container.stop()

    def test_restart(self):
        """A container is restarted."""
        an_container = models.Container.get(self.client, 'an-container')

        an_container.restart()

    def test_freeze(self):
        """A container is suspended."""
        an_container = models.Container.get(self.client, 'an-container')

        an_container.freeze()

    def test_unfreeze(self):
        """A container is resumed."""
        an_container = models.Container.get(self.client, 'an-container')

        an_container.unfreeze()


class TestContainerSnapshots(testing.PyLXDTestCase):
    """Tests for pylxd.models.Container.snapshots."""

    def setUp(self):
        super(TestContainerSnapshots, self).setUp()
        self.container = models.Container.get(self.client, 'an-container')

    def test_get(self):
        """Return a specific snapshot."""
        snapshot = self.container.snapshots.get('an-snapshot')

        self.assertEqual('an-snapshot', snapshot.name)

    def test_all(self):
        """Return all snapshots."""
        snapshots = self.container.snapshots.all()

        self.assertEqual(1, len(snapshots))
        self.assertEqual('an-snapshot', snapshots[0].name)
        self.assertEqual(self.client, snapshots[0].client)
        self.assertEqual(self.container, snapshots[0].container)

    def test_create(self):
        """Create a snapshot."""
        snapshot = self.container.snapshots.create(
            'an-snapshot', stateful=True, wait=True)

        self.assertEqual('an-snapshot', snapshot.name)


class TestSnapshot(testing.PyLXDTestCase):
    """Tests for pylxd.models.Snapshot."""

    def setUp(self):
        super(TestSnapshot, self).setUp()
        self.container = models.Container.get(self.client, 'an-container')

    def test_rename(self):
        """A snapshot is renamed."""
        snapshot = models.Snapshot(
            self.client, container=self.container,
            name='an-snapshot')

        snapshot.rename('an-renamed-snapshot', wait=True)

        self.assertEqual('an-renamed-snapshot', snapshot.name)

    def test_delete(self):
        """A snapshot is deleted."""
        snapshot = models.Snapshot(
            self.client, container=self.container,
            name='an-snapshot')

        snapshot.delete(wait=True)

        # TODO: add an assertion here

    def test_delete_failure(self):
        """If the response indicates delete failure, raise an exception."""
        def not_found(request, context):
            context.status_code = 404
            return json.dumps({
                'type': 'error',
                'error': 'Not found',
                'error_code': 404})
        self.add_rule({
            'text': not_found,
            'method': 'DELETE',
            'url': r'^http://pylxd.test/1.0/containers/an-container/snapshots/an-snapshot$',  # NOQA
        })

        snapshot = models.Snapshot(
            self.client, container=self.container,
            name='an-snapshot')

        self.assertRaises(exceptions.LXDAPIException, snapshot.delete)

    def test_publish(self):
        """Snapshots can be published."""
        self.add_rule({
            'text': json.dumps({
                'type': 'sync',
                'metadata': {
                    'id': 'operation-abc',
                    'metadata': {
                        'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'  # NOQA
                        }
                    }
                }),
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/operations/operation-abc$',
        })

        snapshot = models.Snapshot(
            self.client, container=self.container,
            name='an-snapshot')

        image = snapshot.publish(wait=True)

        self.assertEqual(
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            image.fingerprint)


class TestFiles(testing.PyLXDTestCase):
    """Tests for pylxd.models.Container.files."""

    def setUp(self):
        super(TestFiles, self).setUp()
        self.container = models.Container.get(self.client, 'an-container')

    def test_put(self):
        """A file is put on the container."""
        data = 'The quick brown fox'

        self.container.files.put('/tmp/putted', data)

        # TODO: Add an assertion here

    def test_get(self):
        """A file is retrieved from the container."""
        data = self.container.files.get('/tmp/getted')

        self.assertEqual(b'This is a getted file', data)

    def test_get_not_found(self):
        """LXDAPIException is raised on bogus filenames."""
        def not_found(request, context):
            context.status_code = 500
        rule = {
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-container/files\?path=%2Ftmp%2Fgetted$',  # NOQA
        }
        self.add_rule(rule)

        self.assertRaises(
            exceptions.LXDAPIException,
            self.container.files.get, '/tmp/getted')

    def test_get_error(self):
        """LXDAPIException is raised on error."""
        def not_found(request, context):
            context.status_code = 503
        rule = {
            'text': not_found,
            'method': 'GET',
            'url': r'^http://pylxd.test/1.0/containers/an-container/files\?path=%2Ftmp%2Fgetted$',  # NOQA
        }
        self.add_rule(rule)

        self.assertRaises(
            exceptions.LXDAPIException,
            self.container.files.get, '/tmp/getted')



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import print_function
import datetime
import json

from six.moves import urllib

from pylxd.deprecated import base
from pylxd.deprecated import connection
from pylxd.deprecated import exceptions

image_architecture = {
    0: 'Unknown',
    1: 'i686',
    2: 'x86_64',
    3: 'armv7l',
    4: 'aarch64',
    5: 'ppc',
    6: 'ppc64',
    7: 'ppc64le'
}


class LXDImage(base.LXDBase):

    def __init__(self, conn=None):
        self.connection = conn or connection.LXDConnection()

    # list images
    def image_list(self):
        try:
            (state, data) = self.connection.get_object('GET', '/1.0/images')
            return [image.split('/1.0/images/')[-1]
                    for image in data['metadata']]
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    def image_defined(self, image):
        try:
            (state, data) = self.connection.get_object('GET', '/1.0/images/%s'
                                                       % image)
        except exceptions.APIError as ex:
            if ex.status_code == 404:
                return False
            else:
                raise
        else:
            return True

    def image_list_by_key(self, params):
        try:
            (state, data) = self.connection.get_object(
                'GET', '/1.0/images', urllib.parse.urlencode(params))
            return [image.split('/1.0/images/')[-1]
                    for image in data['metadata']]
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    # image info
    def image_info(self, image):
        try:
            (state, data) = self.connection.get_object('GET', '/1.0/images/%s'
                                                       % image)
            image = {
                'image_upload_date': self.get_image_date(image,
                                                         data.get('metadata'),
                                                         'uploaded_at'),
                'image_created_date': self.get_image_date(image,
                                                          data.get('metadata'),
                                                          'created_at'),
                'image_expires_date': self.get_image_date(image,
                                                          data.get('metadata'),
                                                          'expires_at'),
                'image_public': self.get_image_permission(
                    image,
                    data.get('metadata')),
                'image_size': '%sMB' % self.get_image_size(
                    image,
                    data.get('metadata')),
                'image_fingerprint': self.get_image_fingerprint(
                    image,
                    data.get('metadata')),
                'image_architecture': self.get_image_architecture(
                    image,
                    data.get('metadata')),
            }

            return image
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    def get_image_date(self, image, data, key):
        try:
            if data is None:
                (state, data) = self.connection.get_object(
                    'GET', '/1.0/images/%s' % image)
                data = data.get('metadata')
            if data[key] != 0:
                return datetime.datetime.fromtimestamp(
                    data[key]).strftime('%Y-%m-%d %H:%M:%S')
            else:
                return 'Unknown'
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    def get_image_permission(self, image, data):
        try:
            if data is None:
                (state, data) = self.connection.get_object(
                    'GET', '/1.0/images/%s' % image)
                data = data.get('metadata')
            return True if data['public'] == 1 else False
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    def get_image_size(self, image, data):
        try:
            if data is None:
                (state, data) = self.connection.get_object(
                    'GET', '/1.0/images/%s' % image)
                data = data.get('metadata')
            image_size = data['size']
            if image_size <= 0:
                raise exceptions.ImageInvalidSize()
            return image_size // 1024 ** 2
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    def get_image_fingerprint(self, image, data):
        try:
            if data is None:
                (state, data) = self.connection.get_object(
                    'GET', '/1.0/images/%s' % image)
                data = data.get('metadata')
            return data['fingerprint']
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    def get_image_architecture(self, image, data):
        try:
            if data is None:
                (state, data) = self.connection.get_object(
                    'GET', '/1.0/images/%s' % image)
                data = data.get('metadata')
            return image_architecture[data['architecture']]
        except Exception as e:
            print("Unable to fetch image info - {}".format(e))
            raise

    # image operations
    def image_upload(self, path=None, data=None, headers={}):
        data = data or open(path, 'rb').read()
        try:
            return self.connection.get_object('POST', '/1.0/images',
                                              data, headers)
        except Exception as e:
            print("Unable to upload image - {}".format(e))
            raise

    def image_delete(self, image):
        try:
            return self.connection.get_status('DELETE', '/1.0/images/%s'
                                              % image)
        except Exception as e:
            print("Unable to delete image - {}".format(e))
            raise

    def image_export(self, image):
        try:
            return self.connection.get_raw('GET', '/1.0/images/%s/export'
                                           % image)
        except Exception as e:
            print("Unable to export image - {}".format(e))
            raise

    def image_update(self, image, data):
        try:
            return self.connection.get_status('PUT', '/1.0/images/%s' % image,
                                              json.dumps(data))
        except Exception as e:
            print("Unable to update image - {}".format(e))
            raise

    def image_rename(self, image, data):
        try:
            return self.connection.get_status('POST', '/1.0/images/%s' % image,
                                              json.dumps(data))
        except Exception as e:
            print("Unable to rename image - {}".format(e))
            raise


class LXDAlias(base.LXDBase):

    def alias_list(self):
        (state, data) = self.connection.get_object(
            'GET', '/1.0/images/aliases')
        return [alias.split('/1.0/images/aliases/')[-1]
                for alias in data['metadata']]

    def alias_defined(self, alias):
        return self.connection.get_status('GET', '/1.0/images/aliases/%s'
                                          % alias)

    def alias_show(self, alias):
        return self.connection.get_object('GET', '/1.0/images/aliases/%s'
                                          % alias)

    def alias_update(self, alias, data):
        return self.connection.get_status('PUT',
                                          '/1.0/images/aliases/%s' % alias,
                                          json.dumps(data))

    def alias_rename(self, alias, data):
        return self.connection.get_status('POST',
                                          '/1.0/images/aliases/%s' % alias,
                                          json.dumps(data))

    def alias_create(self, data):
        return self.connection.get_status('POST', '/1.0/images/aliases',
                                          json.dumps(data))

    def alias_delete(self, alias):
        return self.connection.get_status('DELETE', '/1.0/images/aliases/%s'
                                          % alias)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
import time

import six
from six.moves.urllib import parse
try:
    from ws4py.client import WebSocketBaseClient
    from ws4py.manager import WebSocketManager
    _ws4py_installed = True
except ImportError:  # pragma: no cover
    WebSocketBaseClient = object
    _ws4py_installed = False

from pylxd import managers
from pylxd.models import _model as model


class ContainerState(object):
    """A simple object for representing container state."""

    def __init__(self, **kwargs):
        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)


_ContainerExecuteResult = collections.namedtuple(
    'ContainerExecuteResult',
    ['exit_code', 'stdout', 'stderr'])


class Container(model.Model):
    """An LXD Container.

    This class is not intended to be used directly, but rather to be used
    via `Client.containers.create`.
    """

    architecture = model.Attribute()
    config = model.Attribute()
    created_at = model.Attribute()
    devices = model.Attribute()
    ephemeral = model.Attribute()
    expanded_config = model.Attribute()
    expanded_devices = model.Attribute()
    name = model.Attribute(readonly=True)
    profiles = model.Attribute()
    status = model.Attribute(readonly=True)
    last_used_at = model.Attribute(readonly=True)

    status_code = model.Attribute(readonly=True)
    stateful = model.Attribute(readonly=True)

    snapshots = model.Manager()
    files = model.Manager()

    @property
    def api(self):
        return self.client.api.containers[self.name]

    class FilesManager(object):
        """A pseudo-manager for namespacing file operations."""

        def __init__(self, client, container):
            self._client = client
            self._container = container

        def put(self, filepath, data):
            response = self._client.api.containers[
                self._container.name].files.post(
                params={'path': filepath}, data=data)
            return response.status_code == 200

        def get(self, filepath):
            response = self._client.api.containers[
                self._container.name].files.get(
                params={'path': filepath})
            return response.content

    @classmethod
    def exists(cls, client, name):
        """Determine whether a container exists."""
        try:
            client.containers.get(name)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, name):
        """Get a container by name."""
        response = client.api.containers[name].get()

        container = cls(client, **response.json()['metadata'])
        return container

    @classmethod
    def all(cls, client):
        """Get all containers.

        Containers returned from this method will only have the name
        set, as that is the only property returned from LXD. If more
        information is needed, `Container.sync` is the method call
        that should be used.
        """
        response = client.api.containers.get()

        containers = []
        for url in response.json()['metadata']:
            name = url.split('/')[-1]
            containers.append(cls(client, name=name))
        return containers

    @classmethod
    def create(cls, client, config, wait=False):
        """Create a new container config."""
        response = client.api.containers.post(json=config)

        if wait:
            client.operations.wait_for_operation(response.json()['operation'])
        return cls(client, name=config['name'])

    def __init__(self, *args, **kwargs):
        super(Container, self).__init__(*args, **kwargs)

        self.snapshots = managers.SnapshotManager(self.client, self)
        self.files = self.FilesManager(self.client, self)

    def rename(self, name, wait=False):
        """Rename a container."""
        response = self.api.post(json={'name': name})

        if wait:
            self.client.operations.wait_for_operation(
                response.json()['operation'])
        self.name = name

    def _set_state(self, state, timeout=30, force=True, wait=False):
        response = self.api.state.put(json={
            'action': state,
            'timeout': timeout,
            'force': force
        })
        if wait:
            self.client.operations.wait_for_operation(
                response.json()['operation'])
            if 'status' in self.__dirty__:
                del self.__dirty__[self.__dirty__.index('status')]
            self.sync()

    def state(self):
        response = self.api.state.get()
        state = ContainerState(**response.json()['metadata'])
        return state

    def start(self, timeout=30, force=True, wait=False):
        """Start the container."""
        return self._set_state('start',
                               timeout=timeout,
                               force=force,
                               wait=wait)

    def stop(self, timeout=30, force=True, wait=False):
        """Stop the container."""
        return self._set_state('stop',
                               timeout=timeout,
                               force=force,
                               wait=wait)

    def restart(self, timeout=30, force=True, wait=False):
        """Restart the container."""
        return self._set_state('restart',
                               timeout=timeout,
                               force=force,
                               wait=wait)

    def freeze(self, timeout=30, force=True, wait=False):
        """Freeze the container."""
        return self._set_state('freeze',
                               timeout=timeout,
                               force=force,
                               wait=wait)

    def unfreeze(self, timeout=30, force=True, wait=False):
        """Unfreeze the container."""
        return self._set_state('unfreeze',
                               timeout=timeout,
                               force=force,
                               wait=wait)

    def execute(self, commands, environment={}):
        """Execute a command on the container.

        In pylxd 2.2, this method will be renamed `execute` and the existing
        `execute` method removed.
        """
        if not _ws4py_installed:
            raise ValueError(
                'This feature requires the optional ws4py library.')
        if isinstance(commands, six.string_types):
            raise TypeError("First argument must be a list.")
        response = self.api['exec'].post(json={
            'command': commands,
            'environment': environment,
            'wait-for-websocket': True,
            'interactive': False,
        })

        fds = response.json()['metadata']['metadata']['fds']
        operation_id = response.json()['operation'].split('/')[-1]
        parsed = parse.urlparse(
            self.client.api.operations[operation_id].websocket._api_endpoint)

        manager = WebSocketManager()

        stdin = _StdinWebsocket(self.client.websocket_url)
        stdin.resource = '{}?secret={}'.format(parsed.path, fds['0'])
        stdin.connect()
        stdout = _CommandWebsocketClient(manager, self.client.websocket_url)
        stdout.resource = '{}?secret={}'.format(parsed.path, fds['1'])
        stdout.connect()
        stderr = _CommandWebsocketClient(manager, self.client.websocket_url)
        stderr.resource = '{}?secret={}'.format(parsed.path, fds['2'])
        stderr.connect()

        manager.start()

        while len(manager.websockets.values()) > 0:
            time.sleep(.1)

        operation = self.client.operations.get(operation_id)
        return _ContainerExecuteResult(
            operation.metadata['return'], stdout.data, stderr.data)

    def migrate(self, new_client, wait=False):
        """Migrate a container.

        Destination host information is contained in the client
        connection passed in.

        If the container is running, it either must be shut down
        first or criu must be installed on the source and destination
        machines.
        """
        if self.api.scheme in ('http+unix',):
            raise ValueError('Cannot migrate from a local client connection')

        return new_client.containers.create(
            self.generate_migration_data(), wait=wait)

    def generate_migration_data(self):
        """Generate the migration data.

        This method can be used to handle migrations where the client
        connection uses the local unix socket. For more information on
        migration, see `Container.migrate`.
        """
        self.sync()  # Make sure the object isn't stale
        response = self.api.post(json={'migration': True})
        operation = self.client.operations.get(response.json()['operation'])
        operation_url = self.client.api.operations[operation.id]._api_endpoint
        secrets = response.json()['metadata']['metadata']
        cert = self.client.host_info['environment']['certificate']

        return {
            'name': self.name,
            'architecture': self.architecture,
            'config': self.config,
            'devices': self.devices,
            'epehemeral': self.ephemeral,
            'default': self.profiles,
            'source': {
                'type': 'migration',
                'operation': operation_url,
                'mode': 'pull',
                'certificate': cert,
                'secrets': secrets,
            }
        }

    def publish(self, public=False, wait=False):
        """Publish a container as an image.

        The container must be stopped in order publish it as an image. This
        method does not enforce that constraint, so a LXDAPIException may be
        raised if this method is called on a running container.

        If wait=True, an Image is returned.
        """
        data = {
            'public': public,
            'source': {
                'type': 'container',
                'name': self.name,
            }
        }

        response = self.client.api.images.post(json=data)
        if wait:
            operation = self.client.operations.wait_for_operation(
                response.json()['operation'])

            return self.client.images.get(operation.metadata['fingerprint'])


class _CommandWebsocketClient(WebSocketBaseClient):  # pragma: no cover
    def __init__(self, manager, *args, **kwargs):
        self.manager = manager
        super(_CommandWebsocketClient, self).__init__(*args, **kwargs)

    def handshake_ok(self):
        self.manager.add(self)
        self.buffer = []

    def received_message(self, message):
        if len(message.data) == 0:
            self.close()
            self.manager.remove(self)
        if message.encoding:
            self.buffer.append(message.data.decode(message.encoding))
        else:
            self.buffer.append(message.data.decode('utf-8'))

    @property
    def data(self):
        return ''.join(self.buffer)


class _StdinWebsocket(WebSocketBaseClient):  # pragma: no cover
    """A websocket client for handling stdin.

    The nature of stdin in Container.execute means that we don't
    ever use this connection. It is closed as soon as it completes
    the handshake.
    """

    def handshake_ok(self):
        self.close()


class Snapshot(model.Model):
    """A container snapshot."""

    name = model.Attribute()
    stateful = model.Attribute()

    container = model.Parent()

    @property
    def api(self):
        return self.client.api.containers[
            self.container.name].snapshots[self.name]

    @classmethod
    def get(cls, client, container, name):
        response = client.api.containers[
            container.name].snapshots[name].get()

        snapshot = cls(
            client, container=container,
            **response.json()['metadata'])
        # Snapshot names are namespaced in LXD, as
        # container-name/snapshot-name. We hide that implementation
        # detail.
        snapshot.name = snapshot.name.split('/')[-1]
        return snapshot

    @classmethod
    def all(cls, client, container):
        response = client.api.containers[container.name].snapshots.get()

        return [cls(
                client, name=snapshot.split('/')[-1],
                container=container)
                for snapshot in response.json()['metadata']]

    @classmethod
    def create(cls, client, container, name, stateful=False, wait=False):
        response = client.api.containers[container.name].snapshots.post(json={
            'name': name, 'stateful': stateful})

        snapshot = cls(client, container=container, name=name)
        if wait:
            client.operations.wait_for_operation(response.json()['operation'])
        return snapshot

    def rename(self, new_name, wait=False):
        """Rename a snapshot."""
        response = self.api.post(json={'name': new_name})
        if wait:
            self.client.operations.wait_for_operation(
                response.json()['operation'])
        self.name = new_name

    def publish(self, public=False, wait=False):
        """Publish a snapshot as an image.

        If wait=True, an Image is returned.

        This functionality is currently broken in LXD. Please see
        https://github.com/lxc/lxd/issues/2201 - The implementation
        here is mostly a guess. Once that bug is fixed, we can verify
        that this works, or file a bug to fix it appropriately.
        """
        data = {
            'public': public,
            'source': {
                'type': 'snapshot',
                'name': '{}/{}'.format(self.container.name, self.name),
            }
        }

        response = self.client.api.images.post(json=data)
        if wait:
            operation = self.client.operations.wait_for_operation(
                response.json()['operation'])
            return self.client.images.get(operation.metadata['fingerprint'])



import warnings

warnings.simplefilter('once', DeprecationWarning)


class deprecated():  # pragma: no cover
    """A decorator for warning about deprecation warnings.

    The decorator takes an optional message argument. This message can
    be used to direct the user to a new API or specify when it will
    be removed.
    """

    DEFAULT_MESSAGE = '{} is deprecated and will be removed soon.'

    def __init__(self, message=None):
        self.message = message

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            if self.message is None:
                self.message = self.DEFAULT_MESSAGE.format(
                    f.__name__)
            warnings.warn(self.message, DeprecationWarning)
            return f(*args, **kwargs)
        return wrapped



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import contextlib
import tempfile
import uuid
import warnings

import six

from pylxd.models import _model as model


def _image_create_from_config(client, config, wait=False):
    """ Create an image from the given configuration.

    See: https://github.com/lxc/lxd/blob/master/doc/rest-api.md#post-6
    """
    response = client.api.images.post(json=config)
    if wait:
        return client.operations.wait_for_operation(
            response.json()['operation'])
    return response.json()['operation']


class Image(model.Model):
    """A LXD Image."""
    aliases = model.Attribute(readonly=True)
    auto_update = model.Attribute(optional=True)
    architecture = model.Attribute(readonly=True)
    cached = model.Attribute(readonly=True)
    created_at = model.Attribute(readonly=True)
    expires_at = model.Attribute(readonly=True)
    filename = model.Attribute(readonly=True)
    fingerprint = model.Attribute(readonly=True)
    last_used_at = model.Attribute(readonly=True)
    properties = model.Attribute()
    public = model.Attribute()
    size = model.Attribute(readonly=True)
    uploaded_at = model.Attribute(readonly=True)
    update_source = model.Attribute(readonly=True)

    @property
    def api(self):
        return self.client.api.images[self.fingerprint]

    @classmethod
    def exists(cls, client, fingerprint, alias=False):
        """Determine whether an image exists.

        If `alias` is True, look up the image by its alias,
        rather than its fingerprint.
        """
        try:
            if alias:
                client.images.get_by_alias(fingerprint)
            else:
                client.images.get(fingerprint)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, fingerprint):
        """Get an image."""
        response = client.api.images[fingerprint].get()

        image = cls(client, **response.json()['metadata'])
        return image

    @classmethod
    def get_by_alias(cls, client, alias):
        """Get an image by its alias."""
        response = client.api.images.aliases[alias].get()

        fingerprint = response.json()['metadata']['target']
        return cls.get(client, fingerprint)

    @classmethod
    def all(cls, client):
        """Get all images."""
        response = client.api.images.get()

        images = []
        for url in response.json()['metadata']:
            fingerprint = url.split('/')[-1]
            images.append(cls(client, fingerprint=fingerprint))
        return images

    @classmethod
    def create(
            cls, client, image_data, metadata=None, public=False, wait=True):
        """Create an image.

        If metadata is provided, a multipart form data request is formed to
        push metadata and image together in a single request. The metadata must
        be a tar achive.

        `wait` parameter is now ignored, as the image fingerprint cannot be
        reliably determined consistently until after the image is indexed.
        """

        if wait is False:  # pragma: no cover
            warnings.warn(
                'Image.create wait parameter ignored and will be removed in '
                '2.3', DeprecationWarning)

        headers = {}
        if public:
            headers['X-LXD-Public'] = '1'

        if metadata is not None:
            boundary = str(uuid.uuid1())

            data = b''
            for name, contents in (
                    ('metadata', metadata), ('rootfs', image_data)):
                data += b'\r\n'.join([
                    six.b('--{}'.format(boundary)),
                    six.b(
                        'Content-Disposition:form-data;'
                        'name={};filename={}'.format(name, name)),
                    b'Content-Type: application/octet-stream',
                    b'',
                    contents,
                    b'',
                ])
            data += six.b('--{}--\r\n\r\n'.format(boundary))

            headers['Content-Type'] = "multipart/form-data;boundary={}".format(
                boundary)
        else:
            data = image_data

        response = client.api.images.post(data=data, headers=headers)
        operation = client.operations.wait_for_operation(
            response.json()['operation'])
        return cls(client, fingerprint=operation.metadata['fingerprint'])

    @classmethod
    def create_from_simplestreams(cls, client, server, alias,
                                  public=False, auto_update=False):
        """Copy an image from simplestreams."""
        config = {
            'public': public,
            'auto_update': auto_update,

            'source': {
                'type': 'image',
                'mode': 'pull',
                'server': server,
                'protocol': 'simplestreams',
                'fingerprint': alias
            }
        }

        op = _image_create_from_config(client, config, wait=True)

        return client.images.get(op.metadata['fingerprint'])

    @classmethod
    def create_from_url(cls, client, url,
                        public=False, auto_update=False):
        """Copy an image from an url."""
        config = {
            'public': public,
            'auto_update': auto_update,

            'source': {
                'type': 'url',
                'mode': 'pull',
                'url': url
            }
        }

        op = _image_create_from_config(client, config, wait=True)

        return client.images.get(op.metadata['fingerprint'])

    def export(self):
        """Export the image.

        Because the image itself may be quite large, we stream the download
        in 1kb chunks, and write it to a temporary file on disk. Once that
        file is closed, it is deleted from the disk.
        """
        on_disk = tempfile.TemporaryFile()
        with contextlib.closing(self.api.export.get(stream=True)) as response:
            for chunk in response.iter_content(chunk_size=1024):
                on_disk.write(chunk)
        on_disk.seek(0)
        return on_disk

    def add_alias(self, name, description):
        """Add an alias to the image."""
        self.client.api.images.aliases.post(json={
            'description': description,
            'target': self.fingerprint,
            'name': name
        })

        # Update current aliases list
        self.aliases.append({
            'description': description,
            'target': self.fingerprint,
            'name': name
        })

    def delete_alias(self, name):
        """Delete an alias from the image."""
        self.client.api.images.aliases[name].delete()

        # Update current aliases list
        la = [a['name'] for a in self.aliases]
        try:
            del self.aliases[la.index(name)]
        except ValueError:
            pass

    def copy(self, new_client, public=None, auto_update=None, wait=False):
        """Copy an image to a another LXD.

        Destination host information is contained in the client
        connection passed in.
        """
        self.sync()  # Make sure the object isn't stale

        url = '/'.join(self.client.api._api_endpoint.split('/')[:-1])

        if public is None:
            public = self.public

        if auto_update is None:
            auto_update = self.auto_update

        config = {
            'filename': self.filename,
            'public': public,
            'auto_update': auto_update,
            'properties': self.properties,

            'source': {
                'type': 'image',
                'mode': 'pull',
                'server': url,
                'protocol': 'lxd',
                'fingerprint': self.fingerprint
            }
        }

        if self.public is not True:
            response = self.api.secret.post(json={})
            secret = response.json()['metadata']['metadata']['secret']
            config['source']['secret'] = secret
            cert = self.client.host_info['environment']['certificate']
            config['source']['certificate'] = cert

        _image_create_from_config(new_client, config, wait)

        if wait:
            return new_client.images.get(self.fingerprint)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

from pylxd import models
from pylxd.tests import testing


class TestCertificate(testing.PyLXDTestCase):
    """Tests for pylxd.models.Certificate."""

    def test_get(self):
        """A certificate is retrieved."""
        cert = self.client.certificates.get('an-certificate')

        self.assertEqual('certificate-content', cert.certificate)

    def test_all(self):
        """A certificates are returned."""
        certs = self.client.certificates.all()

        self.assertIn('an-certificate', [c.fingerprint for c in certs])

    def test_create(self):
        """A certificate is created."""
        cert_data = open(os.path.join(
            os.path.dirname(__file__), '..', 'lxd.crt')).read().encode('utf-8')
        an_certificate = self.client.certificates.create(
            'test-password', cert_data)

        self.assertEqual(
            'eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c',
            an_certificate.fingerprint)

    def test_fetch(self):
        """A partial object is fully fetched."""
        an_certificate = models.Certificate(
            self.client, fingerprint='an-certificate')

        an_certificate.sync()

        self.assertEqual('certificate-content', an_certificate.certificate)

    def test_delete(self):
        """A certificate is deleted."""
        # XXX: rockstar (08 Jun 2016) - This just executes a code path. An
        # assertion should be added.
        an_certificate = models.Certificate(
            self.client, fingerprint='an-certificate')

        an_certificate.delete()



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import pylxd
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from integration.testing import IntegrationTestCase

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class TestClient(IntegrationTestCase):
    """Tests for `Client`."""

    def test_authenticate(self):
        # This is another test with multiple assertions, as it is a test of
        # flow, rather than a single source of functionality.
        client = pylxd.Client('https://127.0.0.1:8443/', verify=False)

        self.assertFalse(client.trusted)

        client.authenticate('password')

        self.assertTrue(client.trusted)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import warnings

import six

from pylxd import exceptions


class Attribute(object):
    """A metadata class for model attributes."""

    def __init__(self, validator=None, readonly=False, optional=False):
        self.validator = validator
        self.readonly = readonly
        self.optional = optional


class Manager(object):
    """A manager declaration.

    This class signals to the model that it will have a Manager
    attribute.
    """


class Parent(object):
    """A parent declaration.

    Child managers must keep a reference to their parent.
    """


class ModelType(type):
    """A Model metaclass.

    This metaclass converts the declarative Attribute style
    to attributes on the model instance itself.
    """

    def __new__(cls, name, bases, attrs):
        if '__slots__' in attrs and name != 'Model':  # pragma: no cover
            raise TypeError('__slots__ should not be specified.')
        attributes = {}
        for_removal = []
        managers = []

        for key, val in attrs.items():
            if type(val) == Attribute:
                attributes[key] = val
                for_removal.append(key)
            if type(val) in (Manager, Parent):
                managers.append(key)
                for_removal.append(key)
        for key in for_removal:
            del attrs[key]

        slots = list(attributes.keys())
        if '__slots__' in attrs:
            slots = slots + attrs['__slots__']
        for base in bases:
            if '__slots__' in dir(base):
                slots = slots + base.__slots__
        if len(managers) > 0:
            slots = slots + managers
        attrs['__slots__'] = slots
        attrs['__attributes__'] = attributes

        return super(ModelType, cls).__new__(cls, name, bases, attrs)


@six.add_metaclass(ModelType)
class Model(object):
    """A Base LXD object model.

    Objects fetched from the LXD API have state, which allows
    the objects to be used transactionally, with E-tag support,
    and be smart about I/O.

    The model lifecycle is this: A model's get/create methods will
    return an instance. That instance may or may not be a partial
    instance. If it is a partial instance, `sync` will be called
    and the rest of the object retrieved from the server when
    un-initialized attributes are read. When attributes are modified,
    the instance is marked as dirty. `save` will save the changes
    to the server.
    """
    NotFound = exceptions.NotFound
    __slots__ = ['client', '__dirty__']

    def __init__(self, client, **kwargs):
        self.__dirty__ = set()
        self.client = client

        for key, val in kwargs.items():
            try:
                setattr(self, key, val)
            except AttributeError:
                warnings.warn(
                    'Attempted to set unknown attribute "{}" '
                    'on instance of "{}"'.format(
                        key, self.__class__.__name__
                    ))
        self.__dirty__.clear()

    def __getattribute__(self, name):
        try:
            return super(Model, self).__getattribute__(name)
        except AttributeError:
            if name in self.__attributes__:
                self.sync()
                return super(Model, self).__getattribute__(name)
            else:
                raise

    def __setattr__(self, name, value):
        if name in self.__attributes__:
            attribute = self.__attributes__[name]

            if attribute.validator is not None:
                if attribute.validator is not type(value):
                    value = attribute.validator(value)
            self.__dirty__.add(name)
        return super(Model, self).__setattr__(name, value)

    @property
    def dirty(self):
        return len(self.__dirty__) > 0

    def sync(self, rollback=False):
        """Sync from the server.

        When collections of objects are retrieved from the server, they
        are often partial objects. The full object must be retrieved before
        it can modified. This method is called when getattr is called on
        a non-initaliazed object.
        """
        # XXX: rockstar (25 Jun 2016) - This has the potential to step
        # on existing attributes.
        response = self.api.get()
        for key, val in response.json()['metadata'].items():
            if key not in self.__dirty__ or rollback:
                setattr(self, key, val)
                self.__dirty__.remove(key)
        if rollback:
            self.__dirty__.clear()

    def rollback(self):
        """Reset the object from the server."""
        return self.sync(rollback=True)

    def save(self, wait=False):
        """Save data to the server.

        This method should write the new data to the server via marshalling.
        It should be a no-op when the object is not dirty, to prevent needless
        I/O.
        """
        marshalled = self.marshall()
        response = self.api.put(json=marshalled)

        if response.json()['type'] == 'async' and wait:
            self.client.operations.wait_for_operation(
                response.json()['operation'])
        self.__dirty__.clear()

    def delete(self, wait=False):
        """Delete an object from the server."""
        response = self.api.delete()

        if response.json()['type'] == 'async' and wait:
            self.client.operations.wait_for_operation(
                response.json()['operation'])
        self.client = None

    def marshall(self):
        """Marshall the object in preparation for updating to the server."""
        marshalled = {}
        for key, val in self.__attributes__.items():
            if ((not val.readonly and not val.optional) or
                    (val.optional and hasattr(self, key))):
                marshalled[key] = getattr(self, key)
        return marshalled



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from pylxd.deprecated import base
from pylxd.deprecated import exceptions


class LXDContainer(base.LXDBase):
    # containers:

    def container_list(self):
        (state, data) = self.connection.get_object('GET', '/1.0/containers')
        return [container.split('/1.0/containers/')[-1]
                for container in data['metadata']]

    def container_running(self, container):
        (state, data) = self.connection.get_object(
            'GET',
            '/1.0/containers/%s/state' % container)
        data = data.get('metadata')
        container_running = False
        if data['status'].upper() in ['RUNNING', 'STARTING', 'FREEZING',
                                      'FROZEN', 'THAWED']:
            container_running = True
        return container_running

    def container_init(self, container):
        return self.connection.get_object('POST', '/1.0/containers',
                                          json.dumps(container))

    def container_update(self, container, config):
        return self.connection.get_object('PUT', '/1.0/containers/%s'
                                          % container, json.dumps(config))

    def container_defined(self, container):
        _, data = self.connection.get_object('GET', '/1.0/containers')
        try:
            containers = data["metadata"]
        except KeyError:
            raise exceptions.PyLXDException("no metadata in GET containers?")

        container_url = "/1.0/containers/%s" % container
        for ct in containers:
            if ct == container_url:
                return True
        return False

    def container_state(self, container):
        return self.connection.get_object(
            'GET', '/1.0/containers/%s/state' % container)

    def container_start(self, container, timeout):
        action = {'action': 'start', 'force': True, 'timeout': timeout}
        return self.connection.get_object('PUT', '/1.0/containers/%s/state'
                                          % container,
                                          json.dumps(action))

    def container_stop(self, container, timeout):
        action = {'action': 'stop', 'force': True, 'timeout': timeout}
        return self.connection.get_object('PUT', '/1.0/containers/%s/state'
                                          % container,
                                          json.dumps(action))

    def container_suspend(self, container, timeout):
        action = {'action': 'freeze', 'force': True, 'timeout': timeout}
        return self.connection.get_object('PUT', '/1.0/containers/%s/state'
                                          % container,
                                          json.dumps(action))

    def container_resume(self, container, timeout):
        action = {'action': 'unfreeze', 'force': True, 'timeout': timeout}
        return self.connection.get_object('PUT', '/1.0/containers/%s/state'
                                          % container,
                                          json.dumps(action))

    def container_reboot(self, container, timeout):
        action = {'action': 'restart', 'force': True, 'timeout': timeout}
        return self.connection.get_object('PUT', '/1.0/containers/%s/state'
                                          % container,
                                          json.dumps(action))

    def container_destroy(self, container):
        return self.connection.get_object('DELETE', '/1.0/containers/%s'
                                          % container)

    def get_container_log(self, container):
        (state, data) = self.connection.get_object(
            'GET', '/1.0/containers/%s?log=true' % container)
        return data['metadata']['log']

    def get_container_config(self, container):
        (state, data) = self.connection.get_object(
            'GET', '/1.0/containers/%s?log=false' % container)
        return data['metadata']

    def get_container_websocket(self, container):
        return self.connection.get_status(
            'GET',
            '/1.0/operations/%s/websocket?secret=%s'
            % (container['operation'], container['fs']))

    def container_info(self, container):
        (state, data) = self.connection.get_object(
            'GET', '/1.0/containers/%s/state' % container)
        return data['metadata']

    def container_migrate(self, container):
        action = {'migration': True}
        return self.connection.get_object(
            'POST', '/1.0/containers/%s' % container,
            json.dumps(action))

    def container_migrate_sync(self, operation_id, container_secret):
        return self.connection.get_ws(
            '/1.0/operations/%s/websocket?secret=%s'
            % (operation_id, container_secret))

    def container_local_copy(self, container):
        return self.connection.get_object(
            'POST',
            '/1.0/containers', json.dumps(container))

    def container_local_move(self, instance, config):
        return self.connection.get_object(
            'POST',
            '/1.0/containers/%s' % instance, json.dumps(config))

    # file operations
    def get_container_file(self, container, filename):
        return self.connection.get_raw(
            'GET',
            '/1.0/containers/%s/files?path=%s' % (container, filename))

    def put_container_file(self, container, src_file,
                           dst_file, uid, gid, mode):
        with open(src_file, 'rb') as f:
            data = f.read()
        return self.connection.get_object(
            'POST',
            '/1.0/containers/%s/files?path=%s' % (container, dst_file),
            body=data,
            headers={'X-LXD-uid': uid, 'X-LXD-gid': gid, 'X-LXD-mode': mode})

    def container_publish(self, container):
        return self.connection.get_object('POST', '/1.0/images',
                                          json.dumps(container))

    # misc operations
    def run_command(self, container, args, interactive, web_sockets, env):
        env = env or {}
        data = {'command': args,
                'interactive': interactive,
                'wait-for-websocket': web_sockets,
                'environment': env}
        return self.connection.get_object('POST', '/1.0/containers/%s/exec'
                                          % container, json.dumps(data))

    # snapshots
    def snapshot_list(self, container):
        (state, data) = self.connection.get_object(
            'GET',
            '/1.0/containers/%s/snapshots' % container)
        return [snapshot.split('/1.0/containers/%s/snapshots/%s/'
                               % (container, container))[-1]
                for snapshot in data['metadata']]

    def snapshot_create(self, container, config):
        return self.connection.get_object('POST',
                                          '/1.0/containers/%s/snapshots'
                                          % container,
                                          json.dumps(config))

    def snapshot_info(self, container, snapshot):
        return self.connection.get_object('GET',
                                          '/1.0/containers/%s/snapshots/%s'
                                          % (container, snapshot))

    def snapshot_rename(self, container, snapshot, config):
        return self.connection.get_object('POST',
                                          '/1.0/containers/%s/snapshots/%s'
                                          % (container, snapshot),
                                          json.dumps(config))

    def snapshot_delete(self, container, snapshot):
        return self.connection.get_object('DELETE',
                                          '/1.0/containers/%s/snapshots/%s'
                                          % (container, snapshot))



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from pylxd.models import _model as model


class Profile(model.Model):
    """A LXD profile."""

    name = model.Attribute(readonly=True)
    description = model.Attribute()
    config = model.Attribute()
    devices = model.Attribute()

    @classmethod
    def exists(cls, client, name):
        """Determine whether a profile exists."""
        try:
            client.profiles.get(name)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, name):
        """Get a profile."""
        response = client.api.profiles[name].get()
        return cls(client, **response.json()['metadata'])

    @classmethod
    def all(cls, client):
        """Get all profiles."""
        response = client.api.profiles.get()

        profiles = []
        for url in response.json()['metadata']:
            name = url.split('/')[-1]
            profiles.append(cls(client, name=name))
        return profiles

    @classmethod
    def create(cls, client, name, config=None, devices=None):
        """Create a profile."""
        profile = {'name': name}
        if config is not None:
            profile['config'] = config
        if devices is not None:
            profile['devices'] = devices
        client.api.profiles.post(json=profile)
        return cls.get(client, name)

    @property
    def api(self):
        return self.client.api.profiles[self.name]

    def rename(self, new_name):
        """Rename the profile."""
        self.api.post(json={'name': new_name})

        return Profile.get(self.client, new_name)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest
import uuid

from pylxd import exceptions
from pylxd.client import Client
from integration.busybox import create_busybox_image


class IntegrationTestCase(unittest.TestCase):
    """A base test case for pylxd integration tests."""

    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        self.client = Client()
        self.lxd = self.client.api

    def generate_object_name(self):
        """Generate a random object name."""
        # Underscores are not allowed in container names.
        test = self.id().split('.')[-1].replace('_', '')
        rando = str(uuid.uuid1()).split('-')[-1]
        return '{}-{}'.format(test, rando)

    def create_container(self):
        """Create a container in lxd."""
        fingerprint, alias = self.create_image()

        name = self.generate_object_name()
        machine = {
            'name': name,
            'architecture': '2',
            'profiles': ['default'],
            'ephemeral': False,
            'config': {'limits.cpu': '2'},
            'source': {'type': 'image',
                       'alias': alias},
        }
        result = self.lxd['containers'].post(json=machine)
        operation_uuid = result.json()['operation'].split('/')[-1]
        result = self.lxd.operations[operation_uuid].wait.get()

        self.addCleanup(self.delete_container, name)
        return name

    def delete_container(self, name, enforce=False):
        """Delete a container in lxd."""
        # enforce is a hack. There's a race somewhere in the delete.
        # To ensure we don't get an infinite loop, let's count.
        count = 0
        try:
            result = self.lxd['containers'][name].delete()
        except exceptions.LXDAPIException as e:
            if e.response.status_code in (400, 404):
                return
            raise
        while enforce and result.status_code == 404 and count < 10:
            try:
                result = self.lxd['containers'][name].delete()
            except exceptions.LXDAPIException as e:
                if e.response.status_code in (400, 404):
                    return
                raise
            count += 1
        try:
            operation_uuid = result.json()['operation'].split('/')[-1]
            result = self.lxd.operations[operation_uuid].wait.get()
        except KeyError:
            pass  # 404 cases are okay.

    def create_image(self):
        """Create an image in lxd."""
        path, fingerprint = create_busybox_image()
        with open(path, 'rb') as f:
            headers = {
                'X-LXD-Public': '1',
                }
            response = self.lxd.images.post(data=f.read(), headers=headers)
        operation_uuid = response.json()['operation'].split('/')[-1]
        self.lxd.operations[operation_uuid].wait.get()

        alias = self.generate_object_name()
        response = self.lxd.images.aliases.post(json={
            'description': '',
            'target': fingerprint,
            'name': alias
            })

        self.addCleanup(self.delete_image, fingerprint)
        return fingerprint, alias

    def delete_image(self, fingerprint):
        """Delete an image in lxd."""
        try:
            self.lxd.images[fingerprint].delete()
        except exceptions.LXDAPIException as e:
            if e.response.status_code == 404:
                return
            raise

    def create_profile(self):
        """Create a profile."""
        name = self.generate_object_name()
        config = {'limits.memory': '1GB'}
        self.lxd.profiles.post(json={
            'name': name,
            'config': config
            })
        return name

    def delete_profile(self, name):
        """Delete a profile."""
        try:
            self.lxd.profiles[name].delete()
        except exceptions.LXDAPIException as e:
            if e.response.status_code == 404:
                return
            raise

    def assertCommon(self, response):
        """Assert common LXD responses.

        LXD responses are relatively standard. This function makes assertions
        to all those standards.
        """
        self.assertEqual(response.status_code, response.json()['status_code'])
        self.assertEqual(
            ['metadata', 'operation', 'status', 'status_code', 'type'],
            sorted(response.json().keys()))



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

from pylxd.deprecated import base


class LXDProfile(base.LXDBase):

    def profile_list(self):
        '''List profiles on the LXD daemon as an array.'''
        (state, data) = self.connection.get_object('GET', '/1.0/profiles')
        return [profiles.split('/1.0/profiles/')[-1]
                for profiles in data['metadata']]

    def profile_create(self, profile):
        '''Create an LXD Profile'''
        return self.connection.get_status('POST', '/1.0/profiles',
                                          json.dumps(profile))

    def profile_show(self, profile):
        '''Display the LXD profile'''
        return self.connection.get_object('GET', '/1.0/profiles/%s'
                                          % profile)

    def profile_defined(self, profile):
        '''Check for an LXD profile'''
        return self.connection.get_status('GET', '/1.0/profiles/%s'
                                          % profile)

    def profile_update(self, profile, config):
        '''Update the LXD profile (not implemented)'''
        return self.connection.get_status('PUT', '/1.0/profiles/%s'
                                          % profile,
                                          json.dumps(config))

    def profile_rename(self, profile, config):
        '''Rename the LXD profile'''
        return self.connection.get_status('POST', '/1.0/profiles/%s'
                                          % profile,
                                          json.dumps(config))

    def profile_delete(self, profile):
        '''Delete the LXD profile'''
        return self.connection.get_status('DELETE', '/1.0/profiles/%s'
                                          % profile)



class LXDAPIException(Exception):
    """A generic exception for representing unexpected LXD API responses.

    LXD API responses are clearly documented, and are either a standard
    return value, and background operation, or an error. This exception
    is raised on an error case, or when the response status code is
    not expected for the response.

    This exception should *only* be raised in cases where the LXD REST
    API has returned something unexpected.
    """

    def __init__(self, response):
        super(LXDAPIException, self).__init__()
        self.response = response

    def __str__(self):
        if self.response.status_code == 200:  # Operation failure
            return self.response.json()['metadata']['err']

        try:
            data = self.response.json()
            return data['error']
        except (ValueError, KeyError):
            pass
        return self.response.content.decode('utf-8')


class NotFound(LXDAPIException):
    """An exception raised when an object is not found."""


class ClientConnectionFailed(Exception):
    """An exception raised when the Client connection fails."""



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ddt import data
from ddt import ddt
import mock

from pylxd.deprecated import connection
from pylxd.deprecated import exceptions

from pylxd.deprecated.tests import annotated_data
from pylxd.deprecated.tests import fake_api
from pylxd.deprecated.tests import LXDAPITestBase


@ddt
@mock.patch.object(connection.LXDConnection, 'get_object',
                   return_value=('200', fake_api.fake_host()))
class LXDAPIHostTestObject(LXDAPITestBase):

    def test_get_host_info(self, ms):
        result = self.lxd.host_info()
        self.assertEqual(result, {
            'lxd_api_compat_level': 1,
            'lxd_trusted_host': True,
            'lxd_backing_fs': 'ext4',
            'lxd_driver': 'lxc',
            'lxd_version': 0.12,
            'lxc_version': '1.1.2',
            'kernel_version': '3.19.0-22-generic',
        })
        ms.assert_called_once_with('GET', '/1.0')

    host_data = (
        ('lxd_api_compat', 1),
        ('lxd_host_trust', True),
        ('lxd_backing_fs', 'ext4'),
        ('lxd_driver', 'lxc'),
        ('lxc_version', '1.1.2'),
        ('lxd_version', 0.12),
        ('kernel_version', '3.19.0-22-generic'),
    )

    @annotated_data(*host_data)
    def test_get_host_data(self, method, expected, ms):
        result = getattr(self.lxd, 'get_' + method)(data=None)
        self.assertEqual(expected, result)
        ms.assert_called_once_with('GET', '/1.0')

    @annotated_data(*host_data)
    def test_get_host_data_fail(self, method, expected, ms):
        ms.side_effect = exceptions.PyLXDException
        result = getattr(self.lxd, 'get_' + method)(data=None)
        self.assertEqual(None, result)
        ms.assert_called_once_with('GET', '/1.0')


@ddt
@mock.patch.object(connection.LXDConnection, 'get_status')
class LXDAPIHostTestStatus(LXDAPITestBase):

    @data(True, False)
    def test_get_host_ping(self, value, ms):
        ms.return_value = value
        self.assertEqual(value, self.lxd.host_ping())
        ms.assert_called_once_with('GET', '/1.0')

    def test_get_host_ping_fail(self, ms):
        ms.side_effect = Exception
        self.assertRaises(exceptions.PyLXDException, self.lxd.host_ping)
        ms.assert_called_once_with('GET', '/1.0')



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from pylxd.models import _model as model


class Network(model.Model):
    """A LXD network."""
    name = model.Attribute()
    type = model.Attribute()
    used_by = model.Attribute()
    config = model.Attribute()
    managed = model.Attribute()

    @classmethod
    def get(cls, client, name):
        """Get a network by name."""
        response = client.api.networks[name].get()

        network = cls(client, **response.json()['metadata'])
        return network

    @classmethod
    def all(cls, client):
        """Get all networks."""
        response = client.api.networks.get()

        networks = []
        for url in response.json()['metadata']:
            name = url.split('/')[-1]
            networks.append(cls(client, name=name))
        return networks

    @property
    def api(self):
        return self.client.api.networks[self.name]

    def save(self, wait=False):
        """Save is not available for networks."""
        raise NotImplementedError('save is not implemented')

    def delete(self):
        """Delete is not available for networks."""
        raise NotImplementedError('delete is not implemented')



import json


def containers_POST(request, context):
    context.status_code = 202
    return json.dumps({
        'type': 'async',
        'operation': 'operation-abc'})


def container_POST(request, context):
    context.status_code = 202
    if not request.json().get('migration', False):
        return {
            'type': 'async',
            'operation': 'operation-abc'}
    else:
        return {
            'type': 'async',
            'operation': 'operation-abc',
            'metadata': {
                'metadata': {
                    '0': 'abc',
                    '1': 'def',
                    'control': 'ghi',
                }
            }
        }


def container_DELETE(request, context):
    context.status_code = 202
    return json.dumps({
        'type': 'async',
        'operation': 'operation-abc'})


def images_POST(request, context):
    context.status_code = 202
    return json.dumps({
        'type': 'async',
        'operation': 'images-create-operation'})


def image_DELETE(request, context):
    context.status_code = 202
    return json.dumps({
        'type': 'async',
        'operation': 'operation-abc'})


def profiles_POST(request, context):
    context.status_code = 200
    return json.dumps({
        'type': 'sync',
        'metadata': {}})


def profile_DELETE(request, context):
    context.status_code = 200
    return json.dumps({
        'type': 'sync',
        'operation': 'operation-abc'})


def snapshot_DELETE(request, context):
    context.status_code = 202
    return json.dumps({
        'type': 'async',
        'operation': 'operation-abc'})


def profile_GET(request, context):
    name = request.path.split('/')[-1]
    return json.dumps({
        'type': 'sync',
        'metadata': {
            'name': name,
            'description': 'An description',
            'config': {},
            'devices': {},
        },
    })


RULES = [
    # General service endpoints
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {'auth': 'trusted',
                         'environment': {
                             'certificate': 'an-pem-cert',
                             }}}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {'auth': 'trusted',
                         'environment': {}}}),
        'method': 'GET',
        'url': r'^http://pylxd2.test/1.0$',
    },



    # Certificates
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': [
                'http://pylxd.test/1.0/certificates/an-certificate',
            ]}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/certificates$',
    },
    {
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/certificates$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {
                'certificate': 'certificate-content',
                'fingerprint': 'eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c',  # NOQA
                'type': 'client',
            }}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/certificates/eaf55b72fc23aa516d709271df9b0116064bf8cfa009cf34c67c33ad32c2320c$',  # NOQA
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {
                'certificate': 'certificate-content',
                'fingerprint': 'an-certificate',
                'type': 'client',
            }}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/certificates/an-certificate$',
    },
    {
        'json': {
            'type': 'sync',
            'metadata': {},
        },
        'status_code': 202,
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/certificates/an-certificate$',
    },


    # Containers
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': [
                'http://pylxd.test/1.0/containers/an-container',
            ]}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/containers$',
    },
    {
        'text': containers_POST,
        'method': 'POST',
        'url': r'^http://pylxd2.test/1.0/containers$',
    },
    {
        'text': containers_POST,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/containers$',
    },
    {
        'json': {
            'type': 'sync',
            'metadata': {
                'name': 'an-container',

                'architecture': "x86_64",
                'config': {
                    'security.privileged': "true",
                },
                'created_at': "1983-06-16T00:00:00-00:00",
                'last_used_at': "1983-06-16T00:00:00-00:00",
                'devices': {
                    'root': {
                        'path': "/",
                        'type': "disk"
                    }
                },
                'ephemeral': True,
                'expanded_config': {
                    'security.privileged': "true",
                },
                'expanded_devices': {
                    'eth0': {
                        'name': "eth0",
                        'nictype': "bridged",
                        'parent': "lxdbr0",
                        'type': "nic"
                    },
                    'root': {
                        'path': "/",
                        'type': "disk"
                    }
                },
                'profiles': [
                    "default"
                ],
                'stateful': False,
                'status': "Running",
                'status_code': 103
            }},
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/containers/an-container$',
    },
    {
        'json': {
            'type': 'sync',
            'metadata': {
                'status': 'Running',
                'status_code': 103,
                'disk': {
                    'root': {
                        'usage': 10,
                    }
                },
                'memory': {
                    'usage': 15,
                    'usage_peak': 20,
                    'swap_usage': 0,
                    'swap_usage_peak': 5,
                },
                'network': {
                    'l0': {
                        'addresses': [
                            {'family': 'inet',
                             'address': '127.0.0.1',
                             'netmask': '8',
                             'scope': 'local'}
                        ],
                    }
                },
                'pid': 69,
                'processes': 100,
            }},
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/containers/an-container/state$',  # NOQA
    },
    {
        'status_code': 202,
        'json': {
            'type': 'async',
            'operation': 'operation-abc'},
        'method': 'PUT',
        'url': r'^http://pylxd.test/1.0/containers/an-container/state$',  # NOQA
    },
    {
        'json': container_POST,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/containers/an-container$',
    },
    {
        'text': json.dumps({
            'type': 'async',
            'operation': 'operation-abc'}),
        'status_code': 202,
        'method': 'PUT',
        'url': r'^http://pylxd.test/1.0/containers/an-container$',
    },
    {
        'text': container_DELETE,
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/containers/an-container$',
    },
    {
        'json': {
            'type': 'async',
            'metadata': {
                'metadata': {
                    'fds': {
                        '0': 'abc',
                        '1': 'def',
                        '2': 'ghi',
                        'control': 'jkl',
                    }
                },
            },
            'operation': 'operation-abc'},
        'status_code': 202,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/containers/an-container/exec$',  # NOQA
    },

    # Container Snapshots
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': [
                '/1.0/containers/an_container/snapshots/an-snapshot',
            ]}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/containers/an-container/snapshots$',  # NOQA
    },
    {
        'text': json.dumps({
            'type': 'async',
            'operation': 'operation-abc'}),
        'status_code': 202,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/containers/an-container/snapshots$',  # NOQA
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {
                'name': 'an_container/an-snapshot',
                'stateful': False,
            }}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/containers/an-container/snapshots/an-snapshot$',  # NOQA
    },
    {
        'text': json.dumps({
            'type': 'async',
            'operation': 'operation-abc'}),
        'status_code': 202,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/containers/an-container/snapshots/an-snapshot$',  # NOQA
    },
    {
        'text': snapshot_DELETE,
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/containers/an-container/snapshots/an-snapshot$',  # NOQA
    },


    # Container files
    {
        'text': 'This is a getted file',
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/containers/an-container/files\?path=%2Ftmp%2Fgetted$',  # NOQA
    },
    {
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/containers/an-container/files\?path=%2Ftmp%2Fputted$',  # NOQA
    },



    # Images
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': [
                'http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
            ]}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/images$',
    },
    {
        'text': images_POST,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/images$',
    },
    {
        'text': images_POST,
        'method': 'POST',
        'url': r'^http://pylxd2.test/1.0/images$',
    },
    {
        'json': {
            'type': 'sync',
            'status': 'Success',
            'status_code': 200,
            'metadata': {
                'name': 'an-alias',
                'description': 'an-alias',
                'target': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
            }
        },
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/images/aliases/an-alias$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {
                'aliases': [
                    {
                        'name': 'an-alias',  # NOQA
                        'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
                    }
                ],
                'architecture': 'x86_64',
                'cached': False,
                'filename': 'a_image.tar.bz2',
                'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
                'public': False,
                'properties': {},
                'size': 1,
                'auto_update': False,
                'created_at': '1983-06-16T02:42:00Z',
                'expires_at': '1983-06-16T02:42:00Z',
                'last_used_at': '1983-06-16T02:42:00Z',
                'uploaded_at': '1983-06-16T02:42:00Z',

            },
        }),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {
                'aliases': [
                    {
                        'name': 'an-alias',  # NOQA
                        'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
                    }
                ],
                'architecture': 'x86_64',
                'cached': False,
                'filename': 'a_image.tar.bz2',
                'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
                'public': False,
                'properties': {},
                'size': 1,
                'auto_update': False,
                'created_at': '1983-06-16T02:42:00Z',
                'expires_at': '1983-06-16T02:42:00Z',
                'last_used_at': '1983-06-16T02:42:00Z',
                'uploaded_at': '1983-06-16T02:42:00Z',

            },
        }),
        'method': 'GET',
        'url': r'^http://pylxd2.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
    },
    {
        'text': json.dumps({'type': 'async', 'operation': 'operation-abc'}),
        'status_code': 202,
        'method': 'PUT',
        'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
    },
    {
        'text': '0' * 2048,
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/export$',  # NOQA
    },
    {
        'text': image_DELETE,
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855$',  # NOQA
    },

    # Image Aliases
    {
        'json': {
            'type': 'sync',
            'status': 'Success',
            'status_code': 200,
            'metadata': {
                'name': 'an-alias',
                'description': 'an-alias',
                'target': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # NOQA
            }
        },
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/images/aliases/an-alias$',
    },
    {
        'json': {
            'type': 'sync',
            'status': 'Success',
            'metadata': None
        },
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/images/aliases$'
    },
    {
        'json': {
            'type': 'sync',
            'status': 'Success',
            'status_code': 200,
            'metadata': None
        },
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/images/aliases/an-alias$'
    },
    {
        'json': {
            'type': 'sync',
            'status': 'Success',
            'status_code': 200,
            'metadata': None
        },
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/images/aliases/b-alias$'
    },

    # Images secret
    {
        'json': {
            'type': 'sync',
            'status': 'Success',
            'status_code': 200,
            'metadata': {
                'metadata': {
                    'secret': 'abcdefg'
                }
            }
        },
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/images/e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/secret$',  # NOQA
    },

    # Networks
    {
        'json': {
            'type': 'sync',
            'metadata': [
                'http://pylxd.test/1.0/networks/lo',
            ]},
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/networks$',
    },
    {
        'json': {
            'type': 'sync',
            'metadata': {
                'name': 'lo',
                'type': 'loopback',
                'used_by': [],
            }},
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/networks/lo$',
    },


    # Profiles
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': [
                'http://pylxd.test/1.0/profiles/an-profile',
            ]}),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/profiles$',
    },
    {
        'text': profiles_POST,
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/profiles$',
    },
    {
        'text': profile_GET,
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile|an-renamed-profile)$',  # NOQA
    },
    {
        'text': json.dumps({'type': 'sync'}),
        'method': 'PUT',
        'url': r'^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile)$',
    },
    {
        'text': json.dumps({'type': 'sync'}),
        'method': 'POST',
        'url': r'^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile)$',
    },
    {
        'text': profile_DELETE,
        'method': 'DELETE',
        'url': r'^http://pylxd.test/1.0/profiles/(an-profile|an-new-profile)$',
    },


    # Operations
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {'id': 'operation-abc', 'metadata': {'return': 0}},
            }),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/operations/operation-abc$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            }),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/operations/operation-abc/wait$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {'id': 'operation-abc'},
            }),
        'method': 'GET',
        'url': r'^http://pylxd2.test/1.0/operations/operation-abc$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            }),
        'method': 'GET',
        'url': r'^http://pylxd2.test/1.0/operations/operation-abc/wait$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {
                'id': 'images-create-operation',
                'metadata': {
                    'fingerprint': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'  # NOQA
                }
            }
        }),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/operations/images-create-operation$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            }),
        'method': 'GET',
        'url': r'^http://pylxd.test/1.0/operations/images-create-operation/wait$',  # NOQA
    },
    {
        'text': json.dumps({
            'type': 'sync',
            'metadata': {'id': 'operation-abc'},
            }),
        'method': 'GET',
        'url': r'^http://pylxd2.test/1.0/operations/images-create-operation$',
    },
    {
        'text': json.dumps({
            'type': 'sync',
            }),
        'method': 'GET',
        'url': r'^http://pylxd2.test/1.0/operations/images-create-operation/wait$',  # NOQA
    }
]



# Copyright (c) 2015 Canonical Ltd
# Copyright (c) 2015 Mirantis inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import copy
import json
import os
import six
import socket
import ssl
import threading

from six.moves import http_client
from six.moves import queue
try:
    from ws4py import client as websocket
except ImportError:
    websocket = None

from pylxd.deprecated import exceptions
from pylxd.deprecated import utils

if hasattr(ssl, 'SSLContext'):
    # For Python >= 2.7.9 and Python 3.x
    if hasattr(ssl, 'PROTOCOL_TLSv1_2'):
        DEFAULT_TLS_VERSION = ssl.PROTOCOL_TLSv1_2
    else:
        DEFAULT_TLS_VERSION = ssl.PROTOCOL_TLSv1
else:
    # For Python 2.6 and <= 2.7.8
    from OpenSSL import SSL
    DEFAULT_TLS_VERSION = SSL.TLSv1_2_METHOD


class UnixHTTPConnection(http_client.HTTPConnection):

    def __init__(self, path, host='localhost', port=None, strict=None,
                 timeout=None):
        if six.PY3:
            http_client.HTTPConnection.__init__(self, host, port=port,
                                                timeout=timeout)
        else:
            http_client.HTTPConnection.__init__(self, host, port=port,
                                                strict=strict,
                                                timeout=timeout)

        self.path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self.path)
        self.sock = sock


class HTTPSConnection(http_client.HTTPConnection):
    default_port = 8443

    def __init__(self, *args, **kwargs):
        http_client.HTTPConnection.__init__(self, *args, **kwargs)

    def connect(self):
        sock = socket.create_connection((self.host, self.port),
                                        self.timeout, self.source_address)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        (cert_file, key_file) = self._get_ssl_certs()
        self.sock = ssl.wrap_socket(sock, certfile=cert_file,
                                    keyfile=key_file,
                                    ssl_version=DEFAULT_TLS_VERSION)

    @staticmethod
    def _get_ssl_certs():
        return (os.path.join(os.environ['HOME'], '.config/lxc/client.crt'),
                os.path.join(os.environ['HOME'], '.config/lxc/client.key'))


_LXDResponse = namedtuple('LXDResponse', ['status', 'body', 'json'])


if websocket is not None:
    class WebSocketClient(websocket.WebSocketBaseClient):
        def __init__(self, url, protocols=None, extensions=None,
                     ssl_options=None, headers=None):
            """WebSocket client that executes into a eventlet green thread."""
            websocket.WebSocketBaseClient.__init__(self, url, protocols,
                                                   extensions,
                                                   ssl_options=ssl_options,
                                                   headers=headers)
            self._th = threading.Thread(
                target=self.run, name='WebSocketClient')
            self._th.daemon = True

            self.messages = queue.Queue()

        def handshake_ok(self):
            """Starts the client's thread."""
            self._th.start()

        def received_message(self, message):
            """Override the base class to store the incoming message."""
            self.messages.put(copy.deepcopy(message))

        def closed(self, code, reason=None):
            # When the connection is closed, put a StopIteration
            # on the message queue to signal there's nothing left
            # to wait for
            self.messages.put(StopIteration)

        def receive(self):
            # If the websocket was terminated and there are no messages
            # left in the queue, return None immediately otherwise the client
            # will block forever
            if self.terminated and self.messages.empty():
                return None
            message = self.messages.get()
            if message is StopIteration:
                return None
            return message


class LXDConnection(object):

    def __init__(self, host=None, port=8443):
        if host:
            self.host = host
            self.port = port
            self.unix_socket = None
        else:
            if 'LXD_DIR' in os.environ:
                self.unix_socket = os.path.join(os.environ['LXD_DIR'],
                                                'unix.socket')
            else:
                self.unix_socket = '/var/lib/lxd/unix.socket'
            self.host, self.port = None, None
        self.connection = None

    def _request(self, *args, **kwargs):
        if self.connection is None:
            self.connection = self.get_connection()
        self.connection.request(*args, **kwargs)
        response = self.connection.getresponse()

        status = response.status
        raw_body = response.read()
        try:
            if six.PY3:
                body = json.loads(raw_body.decode())
            else:
                body = json.loads(raw_body)
        except ValueError:
            body = None

        return _LXDResponse(status, raw_body, body)

    def get_connection(self):
        if self.host:
            return HTTPSConnection(self.host, self.port)
        return UnixHTTPConnection(self.unix_socket)

    def get_object(self, *args, **kwargs):
        response = self._request(*args, **kwargs)

        if not response.json:
            raise exceptions.PyLXDException('Null Data')
        elif response.status == 200 or (
                response.status == 202 and
                response.json.get('status_code') == 100):
            return response.status, response.json
        else:
            utils.get_lxd_error(response.status, response.json)

    def get_status(self, *args, **kwargs):
        response = self._request(*args, **kwargs)

        if not response.json:
            raise exceptions.PyLXDException('Null Data')
        elif response.json.get('error'):
            utils.get_lxd_error(response.status, response.json)
        elif response.status == 200 or (
                response.status == 202 and
                response.json.get('status_code') == 100):
            return True
        return False

    def get_raw(self, *args, **kwargs):
        response = self._request(*args, **kwargs)

        if not response.body:
            raise exceptions.PyLXDException('Null Body')
        elif response.status == 200:
            return response.body
        else:
            raise exceptions.PyLXDException('Failed to get raw response')

    def get_ws(self, path):
        if websocket is None:
            raise ValueError(
                'This feature requires the optional ws4py library.')
        if self.unix_socket:
            connection_string = 'ws+unix://%s' % self.unix_socket
        else:
            connection_string = (
                'wss://%(host)s:%(port)s' % {'host': self.host,
                                             'port': self.port}
            )

        ws = WebSocketClient(connection_string)
        ws.resource = path
        ws.connect()

        return ws



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import os
import os.path

import requests
import requests_unixsocket
from six.moves.urllib import parse
try:
    from ws4py.client import WebSocketBaseClient
    _ws4py_installed = True
except ImportError:  # pragma: no cover
    WebSocketBaseClient = object
    _ws4py_installed = False

from pylxd import exceptions, managers

requests_unixsocket.monkeypatch()


class _APINode(object):
    """An api node object."""

    def __init__(self, api_endpoint, cert=None, verify=True):
        self._api_endpoint = api_endpoint

        if self._api_endpoint.startswith('http+unix://'):
            self.session = requests_unixsocket.Session()
        else:
            self.session = requests.Session()
            self.session.cert = cert
            self.session.verify = verify

    def __getattr__(self, name):
        return self.__class__(
            '{}/{}'.format(self._api_endpoint, name),
            cert=self.session.cert, verify=self.session.verify)

    def __getitem__(self, item):
        return self.__class__(
            '{}/{}'.format(self._api_endpoint, item),
            cert=self.session.cert, verify=self.session.verify)

    def _assert_response(
            self, response, allowed_status_codes=(200,), stream=False):
        """Assert properties of the response.

        LXD's API clearly defines specific responses. If the API
        response is something unexpected (i.e. an error), then
        we need to raise an exception and have the call points
        handle the errors or just let the issue be raised to the
        user.
        """
        if response.status_code not in allowed_status_codes:
            if response.status_code == 404:
                raise exceptions.NotFound(response)
            raise exceptions.LXDAPIException(response)

        # In the case of streaming, we can't validate the json the way we
        # would with normal HTTP responses, so just ignore that entirely.
        if stream:
            return

        try:
            data = response.json()
        except ValueError:
            # Not a JSON response
            return

        if response.status_code == 200:
            # Synchronous request
            try:
                if data['type'] != 'sync':
                    raise exceptions.LXDAPIException(response)
            except KeyError:
                # Missing 'type' in response
                raise exceptions.LXDAPIException(response)

    @property
    def scheme(self):
        return parse.urlparse(self.api._api_endpoint).scheme

    @property
    def netloc(self):
        return parse.urlparse(self.api._api_endpoint).netloc

    def get(self, *args, **kwargs):
        """Perform an HTTP GET."""
        response = self.session.get(
            self._api_endpoint, *args, **kwargs)
        self._assert_response(response, stream=kwargs.get('stream', False))
        return response

    def post(self, *args, **kwargs):
        """Perform an HTTP POST."""
        response = self.session.post(self._api_endpoint, *args, **kwargs)
        # Prior to LXD 2.0.3, successful synchronous requests returned 200,
        # rather than 201.
        self._assert_response(response, allowed_status_codes=(200, 201, 202))
        return response

    def put(self, *args, **kwargs):
        """Perform an HTTP PUT."""
        response = self.session.put(self._api_endpoint, *args, **kwargs)
        self._assert_response(response, allowed_status_codes=(200, 202))
        return response

    def delete(self, *args, **kwargs):
        """Perform an HTTP delete."""
        response = self.session.delete(self._api_endpoint, *args, **kwargs)
        self._assert_response(response, allowed_status_codes=(200, 202))
        return response


class _WebsocketClient(WebSocketBaseClient):
    """A basic websocket client for the LXD API.

    This client is intentionally barebones, and serves
    as a simple default. It simply connects and saves
    all json messages to a messages attribute, which can
    then be read are parsed.
    """

    def handshake_ok(self):
        self.messages = []

    def received_message(self, message):
        json_message = json.loads(message.data.decode('utf-8'))
        self.messages.append(json_message)


class Client(object):
    """Client class for LXD REST API.

    This client wraps all the functionality required to interact with
    LXD, and is meant to be the sole entry point.

    .. attribute:: containers

        Instance of :class:`Client.Containers
        <pylxd.client.Client.Containers>`:

    .. attribute:: images

        Instance of :class:`Client.Images <pylxd.client.Client.Images>`.

    .. attribute:: operations

        Instance of :class:`Client.Operations
        <pylxd.client.Client.Operations>`.

    .. attribute:: profiles

        Instance of :class:`Client.Profiles <pylxd.client.Client.Profiles>`.

    .. attribute:: api

        This attribute provides tree traversal syntax to LXD's REST API for
        lower-level interaction.

        Use the name of the url part as attribute or item of an api object to
        create another api object appended with the new url part name, ie:

            >>> api = Client().api
            # /
            >>> response = api.get()
            # Check status code and response
            >>> print response.status_code, response.json()
            # /containers/test/
            >>> print api.containers['test'].get().json()

    """

    DEFAULT_CERTS = (
        os.path.expanduser('~/.config/lxc/client.crt'),
        os.path.expanduser('~/.config/lxc/client.key'))

    def __init__(self, endpoint=None, version='1.0', cert=None, verify=True):
        self.cert = cert
        if endpoint is not None:
            if endpoint.startswith('/') and os.path.isfile(endpoint):
                self.api = _APINode('http+unix://{}'.format(
                    parse.quote(endpoint, safe='')))
            else:
                # Extra trailing slashes cause LXD to 301
                endpoint = endpoint.rstrip('/')
                if cert is None and (
                        os.path.exists(self.DEFAULT_CERTS[0]) and
                        os.path.exists(self.DEFAULT_CERTS[1])):
                    cert = self.DEFAULT_CERTS
                self.api = _APINode(endpoint, cert=cert, verify=verify)
        else:
            if 'LXD_DIR' in os.environ:
                path = os.path.join(
                    os.environ.get('LXD_DIR'), 'unix.socket')
            else:
                path = '/var/lib/lxd/unix.socket'
            self.api = _APINode('http+unix://{}'.format(
                parse.quote(path, safe='')))
        self.api = self.api[version]

        # Verify the connection is valid.
        try:
            response = self.api.get()
            if response.status_code != 200:
                raise exceptions.ClientConnectionFailed()
            self.host_info = response.json()['metadata']

        except (requests.exceptions.ConnectionError,
                requests.exceptions.InvalidURL):
            raise exceptions.ClientConnectionFailed()

        self.certificates = managers.CertificateManager(self)
        self.containers = managers.ContainerManager(self)
        self.images = managers.ImageManager(self)
        self.networks = managers.NetworkManager(self)
        self.operations = managers.OperationManager(self)
        self.profiles = managers.ProfileManager(self)

    @property
    def trusted(self):
        return self.host_info['auth'] == 'trusted'

    def authenticate(self, password):
        if self.trusted:
            return
        cert = open(self.api.session.cert[0]).read().encode('utf-8')
        self.certificates.create(password, cert)

        # Refresh the host info
        response = self.api.get()
        self.host_info = response.json()['metadata']

    @property
    def websocket_url(self):
        if self.api.scheme in ('http', 'https'):
            host = self.api.netloc
            if self.api.scheme == 'http':
                scheme = 'ws'
            else:
                scheme = 'wss'
        else:
            scheme = 'ws+unix'
            host = parse.unquote(self.api.netloc)
        url = parse.urlunparse((scheme, host, '', '', '', ''))
        return url

    def events(self, websocket_client=None):
        """Get a websocket client for getting events.

        /events is a websocket url, and so must be handled differently than
        most other LXD API endpoints. This method returns
        a client that can be interacted with like any
        regular python socket.

        An optional `websocket_client` parameter can be
        specified for implementation-specific handling
        of events as they occur.
        """
        if not _ws4py_installed:
            raise ValueError(
                'This feature requires the optional ws4py library.')
        if websocket_client is None:
            websocket_client = _WebsocketClient

        client = websocket_client(self.websocket_url)
        parsed = parse.urlparse(self.api.events._api_endpoint)
        client.resource = parsed.path

        return client



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from dateutil.parser import parse as parse_date

from pylxd.deprecated import base


class LXDOperation(base.LXDBase):

    def operation_list(self):
        (state, data) = self.connection.get_object('GET', '/1.0/operations')
        return data['metadata']

    def operation_show(self, operation):
        (state, data) = self.connection.get_object('GET', operation)

        return {
            'operation_create_time':
                self.operation_create_time(operation, data.get('metadata')),
            'operation_update_time':
                self.operation_update_time(operation, data.get('metadata')),
            'operation_status_code':
                self.operation_status_code(operation, data.get('metadata'))
        }

    def operation_info(self, operation):
        return self.connection.get_object('GET', operation)

    def operation_create_time(self, operation, data):
        if data is None:
            (state, data) = self.connection.get_object('GET', operation)
            data = data.get('metadata')
        return parse_date(data['created_at']).strftime('%Y-%m-%d %H:%M:%S')

    def operation_update_time(self, operation, data):
        if data is None:
            (state, data) = self.connection.get_object('GET', operation)
            data = data.get('metadata')
        return parse_date(data['updated_at']).strftime('%Y-%m-%d %H:%M:%S')

    def operation_status_code(self, operation, data):
        if data is None:
            (state, data) = self.connection.get_object('GET', operation)
            data = data.get('metadata')
        return data['status']

    def operation_wait(self, operation, status_code, timeout):
        if timeout == -1:
            return self.connection.get_status(
                'GET', '%s/wait?status_code=%s'
                % (operation, status_code))
        else:
            return self.connection.get_status(
                'GET', '%s/wait?status_code=%s&timeout=%s'
                % (operation, status_code, timeout))

    def operation_stream(self, operation, operation_secret):
        return self.connection.get_ws(
            '%s/websocket?secret=%s' % (operation, operation_secret))

    def operation_delete(self, operation):
        return self.connection.get_status('DELETE', operation)



from pylxd.models.certificate import Certificate  # NOQA
from pylxd.models.container import Container, Snapshot  # NOQA
from pylxd.models.image import Image  # NOQA
from pylxd.models.network import Network  # NOQA
from pylxd.models.operation import Operation  # NOQA
from pylxd.models.profile import Profile  # NOQA




# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import warnings

from pylxd.deprecated import certificate
from pylxd.deprecated import connection
from pylxd.deprecated import container
from pylxd.deprecated import hosts
from pylxd.deprecated import image
from pylxd.deprecated import network
from pylxd.deprecated import operation
from pylxd.deprecated import profiles


class API(object):

    def __init__(self, host=None, port=8443):
        warnings.warn(
            "pylxd.api.API is deprecated. Please use pylxd.Client.",
            DeprecationWarning)
        conn = self.connection = connection.LXDConnection(host=host, port=port)
        self.hosts = hosts.LXDHost(conn)
        self.image = image.LXDImage(conn)
        self.alias = image.LXDAlias(conn)
        self.network = network.LXDNetwork(conn)
        self.operation = operation.LXDOperation(conn)
        self.profiles = profiles.LXDProfile(conn)
        self.certificate = certificate.LXDCertificate(conn)
        self.container = container.LXDContainer(conn)

    # host
    def host_ping(self):
        return self.hosts.host_ping()

    def host_info(self):
        return self.hosts.host_info()

    def get_lxd_api_compat(self, data=None):
        return self.hosts.get_lxd_api_compat(data)

    def get_lxd_host_trust(self, data=None):
        return self.hosts.get_lxd_host_trust(data)

    def get_lxd_backing_fs(self, data=None):
        return self.hosts.get_lxd_backing_fs(data)

    def get_lxd_driver(self, data=None):
        return self.hosts.get_lxd_driver(data)

    def get_lxc_version(self, data=None):
        return self.hosts.get_lxc_version(data)

    def get_lxd_version(self, data=None):
        return self.hosts.get_lxd_version(data)

    def get_kernel_version(self, data=None):
        return self.hosts.get_kernel_version(data)

    def get_host_certificate(self):
        return self.hosts.get_certificate()

    def host_config(self):
        return self.hosts.host_config()

    # images
    def image_list(self):
        return self.image.image_list()

    def image_defined(self, image):
        return self.image.image_defined(image)

    def image_search(self, params):
        return self.image.image_list_by_key(params)

    def image_info(self, image):
        return self.image.image_info(image)

    def image_upload_date(self, image, data=None):
        return self.image.get_image_date(image, data, 'uploaded_at')

    def image_create_date(self, image, data=None):
        return self.image.get_image_date(image, data, 'created_at')

    def image_expire_date(self, image, data=None):
        return self.image.get_image_date(image, data, 'expires_at')

    def image_upload(self, path=None, data=None, headers={}):
        return self.image.image_upload(path=path, data=data, headers=headers)

    def image_delete(self, image):
        return self.image.image_delete(image)

    def image_export(self, image):
        return self.image.image_export(image)

    def image_update(self, image, data):
        return self.image.image_update(image, data)

    def image_rename(self, image, data):
        return self.image.image_rename(image, data)

    # alias
    def alias_list(self):
        return self.alias.alias_list()

    def alias_defined(self, alias):
        return self.alias.alias_defined(alias)

    def alias_create(self, data):
        return self.alias.alias_create(data)

    def alias_update(self, alias, data):
        return self.alias.alias_update(alias, data)

    def alias_show(self, alias):
        return self.alias.alias_show(alias)

    def alias_rename(self, alias, data):
        return self.alias.alias_rename(alias, data)

    def alias_delete(self, alias):
        return self.alias.alias_delete(alias)

    # containers:
    def container_list(self):
        return self.container.container_list()

    def container_defined(self, container):
        return self.container.container_defined(container)

    def container_running(self, container):
        return self.container.container_running(container)

    def container_init(self, container):
        return self.container.container_init(container)

    def container_update(self, container, config):
        return self.container.container_update(container, config)

    def container_state(self, container):
        return self.container.container_state(container)

    def container_start(self, container, timeout):
        return self.container.container_start(container, timeout)

    def container_stop(self, container, timeout):
        return self.container.container_stop(container, timeout)

    def container_suspend(self, container, timeout):
        return self.container.container_suspend(container, timeout)

    def container_resume(self, container, timeout):
        return self.container.container_resume(container, timeout)

    def container_reboot(self, container, timeout):
        return self.container.container_reboot(container, timeout)

    def container_destroy(self, container):
        return self.container.container_destroy(container)

    def get_container_log(self, container):
        return self.container.get_container_log(container)

    def get_container_config(self, container):
        return self.container.get_container_config(container)

    def get_container_websocket(self, container):
        return self.container.get_container_websocket(container)

    def container_info(self, container):
        return self.container.container_info(container)

    def container_local_copy(self, container):
        return self.container.container_local_copy(container)

    def container_local_move(self, instance, container):
        return self.container.container_local_move(instance, container)

    # file operations
    def get_container_file(self, container, filename):
        return self.container.get_container_file(container, filename)

    def container_publish(self, container):
        return self.container.container_publish(container)

    def put_container_file(self, container, src_file,
                           dst_file, uid=0, gid=0, mode=0o644):
        return self.container.put_container_file(
            container, src_file, dst_file, uid, gid, mode)

    # snapshots
    def container_snapshot_list(self, container):
        return self.container.snapshot_list(container)

    def container_snapshot_create(self, container, config):
        return self.container.snapshot_create(container, config)

    def container_snapshot_info(self, container, snapshot):
        return self.container.snapshot_info(container, snapshot)

    def container_snapshot_rename(self, container, snapshot, config):
        return self.container.snapshot_rename(container, snapshot, config)

    def container_snapshot_delete(self, container, snapshot):
        return self.container.snapshot_delete(container, snapshot)

    def container_migrate(self, container):
        return self.container.container_migrate(container)

    def container_migrate_sync(self, operation_id, container_secret):
        return self.container.container_migrate_sync(
            operation_id, container_secret)

    # misc container
    def container_run_command(self, container, args, interactive=False,
                              web_sockets=False, env=None):
        return self.container.run_command(container, args, interactive,
                                          web_sockets, env)

    # certificates
    def certificate_list(self):
        return self.certificate.certificate_list()

    def certificate_show(self, fingerprint):
        return self.certificate.certificate_show(fingerprint)

    def certificate_delete(self, fingerprint):
        return self.certificate.certificate_delete(fingerprint)

    def certificate_create(self, fingerprint):
        return self.certificate.certificate_create(fingerprint)

    # profiles
    def profile_create(self, profile):
        '''Create LXD profile'''
        return self.profiles.profile_create(profile)

    def profile_show(self, profile):
        '''Show LXD profile'''
        return self.profiles.profile_show(profile)

    def profile_defined(self, profile):
        '''Check to see if profile is defined'''
        return self.profiles.profile_defined(profile)

    def profile_list(self):
        '''List LXD profiles'''
        return self.profiles.profile_list()

    def profile_update(self, profile, config):
        '''Update LXD profile'''
        return self.profiles.profile_update(profile, config)

    def profile_rename(self, profile, config):
        '''Rename LXD profile'''
        return self.profiles.profile_rename(profile, config)

    def profile_delete(self, profile):
        '''Delete LXD profile'''
        return self.profiles.profile_delete(profile)

    # lxd operations
    def list_operations(self):
        return self.operation.operation_list()

    def wait_container_operation(self, operation, status_code, timeout):
        return self.operation.operation_wait(operation, status_code, timeout)

    def operation_delete(self, operation):
        return self.operation.operation_delete(operation)

    def operation_info(self, operation):
        return self.operation.operation_info(operation)

    def operation_show_create_time(self, operation, data=None):
        return self.operation.operation_create_time(operation, data)

    def operation_show_update_time(self, operation, data=None):
        return self.operation.operation_update_time(operation, data)

    def operation_show_status(self, operation, data=None):
        return self.operation.operation_status_code(operation, data)

    def operation_stream(self, operation, operation_secret):
        return self.operation.operation_stream(operation, operation_secret)

    # networks
    def network_list(self):
        return self.network.network_list()

    def network_show(self, network):
        return self.network.network_show(network)

    def network_show_name(self, network, data=None):
        return self.network.show_network_name(network, data)

    def network_show_type(self, network, data=None):
        return self.network.show_network_type(network, data)

    def network_show_members(self, network, data=None):
        return self.network.show_network_members(network, data)



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class PyLXDException(Exception):
    pass


class ContainerUnDefined(PyLXDException):
    pass


class UntrustedHost(PyLXDException):
    pass


class ContainerProfileCreateFail(PyLXDException):
    pass


class ContainerProfileDeleteFail(PyLXDException):
    pass


class ImageInvalidSize(PyLXDException):
    pass


class APIError(PyLXDException):

    def __init__(self, error, status_code):
        msg = 'Error %s - %s.' % (status_code, error)
        super(APIError, self).__init__(msg)
        self.status_code = status_code
        self.error = error



# Copyright (c) 2015 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ddt import ddt
import mock

from pylxd.deprecated import connection

from pylxd.deprecated.tests import annotated_data
from pylxd.deprecated.tests import fake_api
from pylxd.deprecated.tests import LXDAPITestBase


@ddt
@mock.patch.object(connection.LXDConnection, 'get_object',
                   return_value=(200, fake_api.fake_network()))
class LXDAPINetworkTest(LXDAPITestBase):

    def test_list_networks(self, ms):
        ms.return_value = ('200', fake_api.fake_network_list())
        self.assertEqual(
            ['lxcbr0'],
            self.lxd.network_list())
        ms.assert_called_with('GET',
                              '/1.0/networks')

    def test_network_show(self, ms):
        self.assertEqual({
            'network_name': 'lxcbr0',
            'network_type': 'bridge',
            'network_members': ['/1.0/containers/trusty-1'],
        }, self.lxd.network_show('lxcbr0'))
        ms.assert_called_with('GET',
                              '/1.0/networks/lxcbr0')

    @annotated_data(
        ('name', 'lxcbr0'),
        ('type', 'bridge'),
        ('members', ['/1.0/containers/trusty-1']),
    )
    def test_network_data(self, method, expected, ms):
        self.assertEqual(
            expected, getattr(self.lxd,
                              'network_show_' + method)('lxcbr0'))
        ms.assert_called_with('GET',
                              '/1.0/networks/lxcbr0')
